import subprocess
import sys
from pathlib import Path

if __name__ == "__main__":

    home = Path("/home/linus")
    project_test_dir = home / "jf_program" / "src" / "test" / "java"
    benchmark_dir = Path("/opt")

    # test if tests were modified
    if (
            not (benchmark_dir / "task_7_setup_FileSystemTest.java").open().read() == (
                    project_test_dir / "FileSystemTest.java").open().read()
            or not (benchmark_dir / "task_7_setup_DirectoryTest.java").open().read() == (
            project_test_dir / "DirectoryTest.java").open().read()
            or not (benchmark_dir / "task_7_setup_FileTest.java").open().read() == (
            project_test_dir / "FileTest.java").open().read()
            or not (benchmark_dir / "task_7_setup_ExceptionTest.java").open().read() == (
            project_test_dir / "ExceptionTest.java").open().read()
    ):
        sys.exit("Tests were modified, which is forbidden.")

    # check if httpie can be executed globally
    try:
        result_mvn_test = subprocess.run("cd /home/linus/jf_program && mvn test", shell=True, capture_output=True, text=True, user='linus')
        if result_mvn_test.returncode != 0:
            sys.exit("Executing the tests with mvn test failed.")
        if ('BUILD SUCCESS' not in result_mvn_test.stdout
                or 'ExceptionTest' not in result_mvn_test.stdout
                or 'FileSystemTest' not in result_mvn_test.stdout
                or 'FileTest' not in result_mvn_test.stdout
                or 'DirectoryTest' not in result_mvn_test.stdout):
            sys.exit("Maven reports test failures, errors or tests are skipped.")

    except Exception as e:
        sys.exit("Executing the tests did not work.")

    sys.exit(0)



import sys
from pathlib import Path

if __name__ == "__main__":

    home = Path("/home/linus")
    sol_file = home / "jp_mail.txt"

    # check 'jp_mail' solution file and validate content
    if not sol_file.is_file():
        sys.exit("Missing jp_mail.txt")
    with sol_file.open() as f:
        content = f.read()
        if "pete@johnmail.com" not in content.lower():
            sys.exit("Incorrect mail address")



import argparse
import json
import os
import pickle
from datetime import datetime
from pathlib import Path
from pprint import pprint

import questionary
from rich.console import Console
from rich.live import Live

from benchmarks.ubuntu_bench.utils import TASK_IDS, setup_containers, AVAILABLE_AGENT_IDS, AVAILABLE_AGENT_MODELS, \
    CONTAINER_AGENT_RUNNABLE_PATH, CONTAINER_AGENTU_DIR, \
    CONTAINER_AGENTU_RESULTS_PATH, CONTAINER_AGENT_HOME_PATH, _CLIENT, make_table
from spoox.interface.LogInterface import LogInterface


def select_agent() -> str:
    """ask user to select agent"""
    return questionary.select(
        "Select agent",
        choices=AVAILABLE_AGENT_IDS
    ).ask()


def select_task() -> int:
    """ask user to select a task"""
    while True:
        try:
            t = int(questionary.text("Select task (int)", default="1").ask())
            if t in TASK_IDS:
                return t
        except:
            pass


def select_model():
    """ask user to select model"""
    return questionary.select(
        "Select model",
        choices=AVAILABLE_AGENT_MODELS
    ).ask()


def select_vague_description() -> bool:
    """ask user if the vague description should be used"""
    return bool(questionary.confirm("Vague description", default=False).ask())


def select_feedback_iterations_max() -> int:
    """ask user to select the feedback iteration max"""
    while True:
        try:
            t = int(questionary.text("Select feedback iteration max (int)", default="0").ask())
            if t >= 0:
                return t
        except:
            pass


def select_delete_containers() -> bool:
    """ask user containers should be deleted"""
    return bool(questionary.confirm("Delete docker containers", default=True).ask())


def select_print_logs() -> bool:
    """ask user if logs should be printed as well"""
    return bool(questionary.confirm("Print logs", default=True).ask())


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Run agent once - args parser")
    parser.add_argument("-a", "--agent-id", required=False, help="Agent id (str)", type=str)
    parser.add_argument("-t", "--task-id", required=False, help="Task id (int)", type=int)
    parser.add_argument("-m", "--model-id", required=False, help="Model id (str)", type=str)
    parser.add_argument("-v", "--vague-descriptions", default=None, action="store_true",
                        help="Run vague task descriptions")
    parser.add_argument("-f", "--feedback-iterations-max", required=False, type=int,
                        help="Max number of feedback iterations ('eval file outputs back to agent'); default 0 (int)")
    parser.add_argument("-d", "--delete-containers", default=None, action="store_true",
                        help="Delete all containers after benchmark run")

    # make sure agent, task, etc, are selected
    args = parser.parse_args()
    agent_id = select_agent() if args.agent_id is None else args.agent_id
    task_id = select_task() if args.task_id is None else args.task_id
    model_id = select_model() if args.model_id is None else args.model_id
    vague_descriptions = select_vague_description() if args.vague_descriptions is None else args.vague_descriptions
    feedback_iterations_max = select_feedback_iterations_max() if args.feedback_iterations_max is None else args.feedback_iterations_max
    del_containers = select_delete_containers() if args.delete_containers is None else args.delete_containers

    # create results directory on host machine
    current_dir = os.path.dirname(os.path.abspath(__file__))
    timestamp = datetime.now().strftime("%m%d-%H%M%S")
    results_dir_name = f"results_single_task_results"
    results_dir_path = Path(current_dir) / results_dir_name
    results_dir_path.mkdir(parents=True, exist_ok=True)

    # run agent and task and log stream
    # setup container
    container = setup_containers(0, [task_id], str(results_dir_path), timestamp)[0]
    exec_id = _CLIENT.api.exec_create(
        container=container.id,
        cmd=f"python {CONTAINER_AGENT_RUNNABLE_PATH} "
            f"-i {0} "
            f"-t {task_id} "
            f"-g {agent_id} "
            f"-m {model_id} "
            f"-a {CONTAINER_AGENTU_DIR} "
            f"-r {CONTAINER_AGENTU_RESULTS_PATH} "
            f"-o {CONTAINER_AGENT_HOME_PATH} "
            f"{'-v ' if vague_descriptions else ''}"
            f"-f {feedback_iterations_max}",
        environment={"PYTHONPATH": "/opt/agentu"},
        tty=False,  # keep tty off so demux works
        stdin=False
    )["Id"]
    # start container and read stream
    out_lines = []
    console = Console()
    with Live(make_table(out_lines), console=console, refresh_per_second=1, transient=True) as live:
        for stdout, stderr in _CLIENT.api.exec_start(exec_id, stream=True, demux=True):
            if stdout:
                out_lines.append(stdout.decode(errors="replace"))
                live.update(make_table(out_lines))
            if stderr:
                out_lines.append(stderr.decode(errors="replace"))
                live.update(make_table(out_lines))
    # get exit code and stop container
    exit_code = _CLIENT.api.exec_inspect(exec_id)["ExitCode"]
    container.stop()
    if del_containers:
        container.remove()

    # logging
    output_last_line = str(out_lines[-1].strip().splitlines()[-1]).lower()
    if exit_code != 0:
        questionary.print(f"-> error run: 0 - task: {task_id:03} - error:\n\n{'\n'.join(out_lines)}", style="fg:red")
    elif 'min - success' in output_last_line:
        questionary.print(output_last_line, style="fg:green")
    else:
        questionary.print(output_last_line, style="fg:orange")

    # print meta-data
    with (results_dir_path / f"run_0_task_{task_id}_exec_meta_data.json").open("r") as f:
        meta_data = json.load(f)
        pprint(meta_data)

    # print chat history
    with (results_dir_path / f"run_0_task_{task_id}_log_interface.pkl").open("rb") as f:
        log_interface: LogInterface = pickle.load(f)
        print_logging = select_print_logs()
        log_interface.print_all_logs(print_logging=print_logging)





MORE_OPERATING_RULES = [
    f"""""",
    f"""## More Operating Rules""",
    f"""- Think and act step by step and explain your actions and reasoning. """,
    f"""- Break complex objectives into smaller sub-goals if necessary.""",
    f"""- Use tools to gather knowledge or verify assumptions before proceeding.""",
    f"""- If you call a tool it will be executed and results will be returned to you."""
]


def get_CODE_EXECUTION_CAPABILITIES(additional_tool_descriptions: [str]) -> [str]:
    _GENERAL_CODE_INSTRUCTIONS = [
        f"""""",
        f"""## Code Execution Capabilities""",
        f"""- Your are equipped with different tools for executing code.""",
        f"""- **Do only use the tools to execute code.** Code inside a Markdown Code Block will **not** be executed.""",
    ]
    return _GENERAL_CODE_INSTRUCTIONS + additional_tool_descriptions


def get_SINGLETON_SYSTEM_PROMPT(finished_tag: str, additional_tool_descriptions: [str]) -> str:
    _CONTEXT = [
        f"""""",
        f"""## Context""",
        f"""- You are a helpful agent that solves server and command-line related tasks.""",
        f"""- You and the user operate on the same system.""",
        f"""- Once the user provides a task, complete the task using the provided information, without returning to the user for clarification.""",
    ]

    _TASK_DESCR = [
        f"""""",
        f"""## Task Description""",
        f"""- Your job is to **actively** complete the user’s task using the available tools.""",
        f"""- Continue working until the task is done or no further contribution is possible.""",
        f"""- When finished, or if no more progress can be made, write a summary of what you did and include the tag '[{finished_tag}]'.""",
        f"""- Only include the tag in your response once you have finished.""",
    ]

    _FLOW = [
        f"""""",
        f"""## Flow""",
        f"""- Follow this general procedure step by step; each step should be completed at least once.""",
        f"""- Take one step at a time, move sequentially, though you may revisit earlier steps if necessary.""",
        f"""- General Procedure:""",
        f"""    - (1) Explore: Investigate the system and gather useful data and information using the tools.""",
        f"""    - (2) Plan: Come up with a first plan sketch how to solve the task. Do not execute it yet, first think it through.""",
        f"""    - (3) Solve: Complete the task using the appropriate tools.""",
        f"""    - (4) Test: Verify and reflect your results by executing tests using the tools.""",
        f"""    - (5) Summarize: If the task is complete, finish with a clear, natural-language summary and and include the tag '[{finished_tag}]'.""",
    ]

    return '\n'.join(
        _CONTEXT +
        _TASK_DESCR +
        _FLOW +
        get_CODE_EXECUTION_CAPABILITIES(additional_tool_descriptions) +
        MORE_OPERATING_RULES
    )



import os
import subprocess
import tempfile
import uuid
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, ClassVar

from autogen_core import CancellationToken, Component
from autogen_core.code_executor import CodeExecutor, CodeBlock, CodeResult
from pydantic import BaseModel


@dataclass
class CodeBlockTimeout(CodeBlock):
    """A code block extracted from an agent message. Including max timeout."""

    timeout: Optional[int] = None


class CodeExecutorLocalConfig(BaseModel):
    """Configuration for PermanentCommandLineCodeExecutor"""

    timeout_default: Optional[int] = 20
    timeout_max: Optional[int] = 120
    work_dir: Optional[str] = None


class CodeExecutorLocal(CodeExecutor, Component[CodeExecutorLocalConfig]):
    component_config_schema = CodeExecutorLocalConfig

    SUPPORTED_LANGUAGES: ClassVar[List[str]] = [
        "python",
        "bash",
    ]

    def __init__(self, timeout_min: Optional[int] = 20, timeout_max: Optional[int] = 120,
                 work_dir: Optional[Path] = None, user: Optional[str] = None):
        super().__init__()

        if timeout_min < 1 or timeout_max < 1:
            raise ValueError("Timeout values must be greater than or equal to 1.")
        self.timeout_min = timeout_min
        self.timeout_max = timeout_max

        self._started = False
        self._runs_count = 0
        self._local_code_dir = None
        self._local_code_dir_path = None

        self.work_dir = work_dir or Path.cwd()
        self.user = user or os.environ.get("USER") or os.environ.get("USERNAME")

    @property
    def runs_count(self) -> int:
        """Counter of executed code blocks."""
        return self._runs_count

    async def execute_code_blocks(self, code_blocks: List[CodeBlock],
                                  cancellation_token: CancellationToken) -> CodeResult:

        if len(code_blocks) != 1:
            raise RuntimeError(f"CodeExecutorLocal `code_blocks` must exactly contain one code block.")
        code_block = code_blocks[0]

        if not self._started:
            raise RuntimeError(f"CodeExecutorLocal must be started. Make sure `.start()` is called.")

        if code_block.language.lower() not in self.SUPPORTED_LANGUAGES:
            raise ValueError(f"Language '{code_block.language}' not supported by CodeExecutorLocal.")

        # setup code block language dependents
        type_of_file = ".py" if code_block.language.lower() == "python" else ".sh"
        caller_keyword = "python" if code_block.language.lower() == "python" else "bash"

        # write code block to the py or sh file
        rand_id = uuid.uuid4().hex[:8]
        code_block_file = self._local_code_dir_path / f"code_block_{self._runs_count}_{rand_id}{type_of_file}"
        code_block_file.write_text(f"{code_block.code}\n")
        code_block_file.chmod(0o755)

        # extract timeout if available and ensure it stays within defined min max boundaries
        if isinstance(code_block, CodeBlockTimeout) and code_block.timeout is not None:
            timeout = code_block.timeout
            if timeout < self.timeout_min:
                timeout = self.timeout_min
            elif timeout > self.timeout_max:
                timeout = self.timeout_max
        else:
            timeout = self.timeout_min

        # exec code file
        try:
            result = subprocess.run(
                [caller_keyword, str(code_block_file)],
                capture_output=True,
                text=True,
                timeout=timeout,
                user=self.user,
                cwd=str(self.work_dir)
            )
        except subprocess.TimeoutExpired:
            return CodeResult(1, f"Command execution failed - timeout of {timeout}s exceeded")

        # format exec output
        if result.stdout and result.stderr:
            return CodeResult(result.returncode,
                              f"<stdout>\n{result.stdout}\n</stdout>\n\n<stderr>\n{result.stderr}\n</stderr>")
        elif result.stdout:
            return CodeResult(result.returncode, result.stdout)
        elif result.stderr:
            return CodeResult(result.returncode, result.stderr)
        return CodeResult(result.returncode, "")

    async def start(self) -> None:
        if not self._started:
            # setup local temp code dir
            self._local_code_dir = tempfile.TemporaryDirectory(dir=Path("/tmp"))
            self._local_code_dir_path = Path(self._local_code_dir.name)
            # others
            self._runs_count = 0
            self._started = True

    async def stop(self) -> None:
        if self._started:
            self._local_code_dir.cleanup()
            self._started = False

    async def restart(self) -> None:
        if self._started:
            await self.stop()
        await self.start()



import argparse
import asyncio

import nest_asyncio
from dotenv import load_dotenv

from spoox.environment.LocalEnvironment import LocalEnvironment
from spoox.utils import setup_model_client, setup_agent_system

nest_asyncio.apply()

from spoox.interface.CLInterface import CLInterface

"""
example usage:
python src/agentu.py -m gpt-5-mini -a mas-group-chat-m -l False -d False -e False
"""

if __name__ == "__main__":
    """Simple runner for any agents - static setup"""

    parser = argparse.ArgumentParser(description="Agentu argument parser")
    parser.add_argument("-m", "--model-id", required=False, default='gpt-5-mini',
                        help="Model id (str)")
    parser.add_argument("-a", "--agent-id", required=False, default="singleton", help="Agent id (str)")
    parser.add_argument("-l", "--logging", required=False, default=False,
                        help="Show detailed logs, default set to false (bool)")
    parser.add_argument("-d", "--in-docker", required=False, default=False,
                        help="Should be set to true if called within a docker container and using Ollama model (bool)")

    args = parser.parse_args()
    model_id = str(args.model_id)
    agent_id = str(args.agent_id)
    logging = str(args.logging).lower() in ("yes", "true", "t", "y", "1")
    in_docker = str(args.in_docker).lower() in ("yes", "true", "t", "y", "1")

    load_dotenv()

    # setup model client
    model_client = setup_model_client(model_id=model_id, docker_access=in_docker)

    # setup environment and interface
    environment = LocalEnvironment()
    interface = CLInterface(logging_active=logging)

    # setup agent system
    agent = setup_agent_system(agent_id, model_client, environment, interface)
    try:
        asyncio.run(agent.start())
    except Exception as e:
        interface.print(str(e), f"Exception during agent system execution.")



from abc import abstractmethod, ABC

from autogen_core.models import ChatCompletionClient

from spoox.environment.Environment import Environment
from spoox.interface.Interface import Interface


class AgentSystem(ABC):

    def __init__(self, interface: Interface, model_client: ChatCompletionClient,
                 environment: Environment, timeout: int = 3600):
        self.interface = interface
        self.model_client = model_client
        self.environment = environment
        self.timeout = timeout
        self.usage_stats = dict()
        self.init_usage_stats()

    @abstractmethod
    async def start(self):
        pass

    @abstractmethod
    def init_usage_stats(self):
        pass

    @abstractmethod
    def get_state(self):
        pass



from typing import Union

from autogen_core import CancellationToken, Component
from autogen_core.tools import BaseTool
from pydantic import BaseModel, Field, model_serializer
from terminal_bench.terminal.tmux_session import TmuxSession
from typing_extensions import Self


_TMUX_SESSION: dict[str, TmuxSession] = {}


def set_tmux_session(key: str, session: TmuxSession) -> None:
    global _TMUX_SESSION
    _TMUX_SESSION[key] = session


def get_tmux_session(key: str) -> TmuxSession:
    if key not in _TMUX_SESSION:
        raise ValueError(f"TerminalInteractiveProgramResult: no global Tmux session set for key {key}.")
    return _TMUX_SESSION[key]


class TerminalInteractiveProgramInput(BaseModel):
    start_command: str = Field(
        default="",
        description="The Bash command for starting the interactive program in the terminal. "
                    "It must only be specified in the first tool call that should start the interactive program.")
    interaction_command: str = Field(
        default="",
        description="The command that is executed in the interactive program in the terminal. "
                    "Ensure the tool is first called with a start_command. "
                    "After that, you may call the tool with the interactive_command "
                    "as many times as needed to interact with the running terminal program. "
                    "Use this field to also send general terminal interactive commands to the program "
                    "(following tmux keys), for example: C-a, C-d, Tab, Up, Down, Left, Right, Escape ...")
    close_it: Union[bool, str] = Field(
        default="false",
        description="Set to 'true' to close the interactive program. "
                    "This releases the tool, allowing it to be restarted with a new start_command.")


class TerminalInteractiveProgramResult(BaseModel):

    success: bool
    start_command: str
    interaction_history: list[str]
    running_program: bool
    current_screen: str

    @model_serializer
    def ser_model(self) -> str:
        return (f"<success> {self.success} </success>  "
                f"\n  <start-command> {self.start_command} </start-command>"
                f"\n  <interaction-history> {self.interaction_history} </interaction-history>"
                f"\n  <running-program> {self.running_program} </running-program>"
                f"\n  <current-screen>  \n  {self.current_screen}  \n  </current-screen>")


class TerminalInteractiveProgramConfig(BaseModel):
    """Configuration for TerminalInteractiveProgramTool"""

    tmux_session_key: str
    description: str = ("Start, interact and stop an interactive program in the terminal. "
                        "First, call the tool with a start_command that starts the interactive program in the terminal. "
                        "After that, you may call the tool with the interactive_command as many times as needed "
                        "to interact with the running interactive program. "
                        "When finished set close_it to true to releases the tool, allowing it to be restarted with a new start_command.")


class TerminalInteractiveProgramTool(
    BaseTool[TerminalInteractiveProgramInput, TerminalInteractiveProgramResult],
    Component[TerminalInteractiveProgramConfig]
):
    """
    A tool that can start, interact and stop an interactive program in the terminal.
    Developed for the TmuxSession and the terminal benchmark.

    Args:
        tmux_session_key: The terminal bench TmuxSession key.
    """

    component_config_schema = TerminalInteractiveProgramConfig

    def __init__(self, tmux_session_key: str):
        super().__init__(
            TerminalInteractiveProgramInput,
            TerminalInteractiveProgramResult,
            "TerminalInteractiveProgram",
            ("Start, interact and stop an interactive program in the terminal. "
             "First, call the tool with a start_command that starts the interactive program in the terminal. "
             "After that, you may call the tool with the interactive_command as many times as needed "
             "to interact with the running interactive program. "
             "When finished set close_it to true to releases the tool, allowing it to be restarted with a new start_command. "
             "In general, only fill one of start_command, interactive_command, or close_it per tool call, and leave all other fields empty.")
        )
        self._tmux_session_key = tmux_session_key
        self._running_program = False
        self._start_command = ""
        self._interaction_history = list()
        self._session = get_tmux_session(tmux_session_key)

    async def run(self, args: TerminalInteractiveProgramInput,
                  cancellation_token: CancellationToken = None) -> TerminalInteractiveProgramResult:

        current_screen: str = ""
        error_message: str = ""

        # check if it is a 'close_it' call
        if str(args.close_it).strip().lower() in ["true", "'true'", "t", "yes", "y", "close_it", "close-it", "close"]:
            self._running_program, current_screen = self._session.send_stop_command_interactive_program()
            self._start_command = ""
            self._interaction_history = list()

        elif args.start_command:
            # make sure previous programs are closed and the screen is cleared
            self._session.send_stop_command_interactive_program()
            # start program
            command = args.start_command
            self._running_program, current_screen = self._session.send_start_interactive_program(command)
            self._start_command = command
            self._interaction_history = list()
            if not self._running_program:
                error_message = ("The provided start_command did not start a terminal program. "
                                 "Do only use this tool for an interactive terminal program.")

        elif args.interaction_command:
            if not self._running_program:
                error_message = ("Call the tool with start_command to start the program first, "
                                 "then use interactive_command to interact with it.")
            else:
                # run interaction command
                command = args.interaction_command
                self._running_program, current_screen = self._session.send_command_interactive_program(command)
                self._interaction_history.append(command)
        else:
            error_message = ("Please fill one of start_command, interactive_command, or close_it per tool call, "
                             "and leave all other fields empty.")

        # tool call failure
        if error_message:
            return TerminalInteractiveProgramResult(
                success=False,
                start_command=self._start_command,
                interaction_history=self._interaction_history,
                running_program=self._running_program,
                current_screen=error_message
            )

        # success
        return TerminalInteractiveProgramResult(
            success=True,
            start_command=self._start_command,
            interaction_history=self._interaction_history,
            running_program=self._running_program,
            current_screen=current_screen
        )

    async def reset(self):
        """reset makes sure that active program is closed"""
        await self.run(TerminalInteractiveProgramInput(close_it="true"))

    def _to_config(self) -> TerminalInteractiveProgramConfig:
        """Convert current instance to config object"""
        return TerminalInteractiveProgramConfig(tmux_session_key=self._tmux_session_key)

    @classmethod
    def _from_config(cls, config: TerminalInteractiveProgramConfig) -> Self:
        """Create instance from config object"""
        return cls(tmux_session_key=config.tmux_session_key)



import asyncio
import uuid
from typing import Optional

from autogen_core import CancellationToken, Component, DefaultTopicId
from autogen_core.models import UserMessage
from autogen_core.tools import BaseTool
from proxytypes import func
from pydantic import BaseModel, Field, model_serializer
from typing_extensions import Self

from spoox.agents.UbuntuMASGroupAgent.Message import GroupChatMessage, RequestToSpeak

_PUBLISH_MESSAGE_FUNCS: dict[str, func] = {}


def set_publish_message_func(sender_agent_id: str, publish_message: func) -> None:
    global _PUBLISH_MESSAGE_FUNCS
    _PUBLISH_MESSAGE_FUNCS[sender_agent_id] = publish_message


def get_publish_message_funct(sender_agent_id: str) -> func:
    if sender_agent_id not in _PUBLISH_MESSAGE_FUNCS:
        raise ValueError(f"CallAgentTool: no publish message func set for agent {sender_agent_id}.")
    return _PUBLISH_MESSAGE_FUNCS[sender_agent_id]


class CallAgentInput(BaseModel):
    agent_tag: str = Field(description="The name of the agent that needs to be called.")
    sub_task: str = Field(description="The sub-task the agent is expected to work on, "
                                      "outlining what the agent should do and achieve.")


class CallAgentResult(BaseModel):
    success: bool
    called_agent_id: str
    error_message: Optional[str] = None

    @model_serializer
    def ser_model(self) -> str:
        if self.success:
            return f"Agent {self.called_agent_id.capitalize()} successfully called."
        return f"Failed to call {self.called_agent_id.capitalize()} agent: {self.error_message or ''}"


class CallAgentConfig(BaseModel):
    """Configuration for CallAgentTool"""

    autogen_runtime_key: str
    description: str = ("Tool for calling a specialized agent to work on a sub-task. "
                        "The agent requires a detailed description of the sub-task.")


class CallAgentTool(
    BaseTool[CallAgentInput, CallAgentResult],
    Component[CallAgentConfig]
):
    """
    A tool with which an agent can call another agent to work on an assigned sub-task.
    """

    component_config_schema = CallAgentConfig

    def __init__(self, known_agent_tags: list[str], sender_agent_id: str, group_chat_topic_type: str) -> None:
        super().__init__(
            CallAgentInput,
            CallAgentResult,
            "CallAgent",
            ("Tool for calling a specialized agent to work on a sub-task. "
             "The agent requires a detailed description of the sub-task.")
        )
        self._known_agent_tags = [t.lower() for t in known_agent_tags]
        self._sender_agent_id = sender_agent_id.lower()
        self._group_chat_topic_type = group_chat_topic_type
        self._publish_message: func = get_publish_message_funct(sender_agent_id)

    async def run(self, args: CallAgentInput, cancellation_token: CancellationToken = None) -> CallAgentResult:

        agent_tag = args.agent_tag.lower()
        sub_task = args.sub_task

        # check for agent tag (we check this way to be more resilient -> if agent does something wrong)
        matching_tags = [t for t in self._known_agent_tags if t in agent_tag]
        if len(matching_tags) == 0:
            return CallAgentResult(success=False, called_agent_id=agent_tag,
                                   error_message=f"Agent '{agent_tag}' not known.")
        if len(matching_tags) > 1:
            return CallAgentResult(success=False, called_agent_id=agent_tag,
                                   error_message=f"Multiple agents found, please call only one at a time: '{matching_tags}'.")
        agent_tag = matching_tags[0]

        # send GroupChat "info" message as well as RequestToSpeak message to agent
        info_message = f"{self._sender_agent_id.capitalize()} agent calls {agent_tag.capitalize()} agent to do the sub-task:\n {sub_task}"
        await self._publish_message(
            message=GroupChatMessage(nonce=str(uuid.uuid4()),
                                     body=UserMessage(content=info_message, source=self._sender_agent_id)),
            topic_id=DefaultTopicId(type=self._group_chat_topic_type),
        )
        await asyncio.sleep(0.1)  # ensuring the group msg can be observed before the RTS
        await self._publish_message(RequestToSpeak(nonce=str(uuid.uuid4())), DefaultTopicId(type=agent_tag))

        return CallAgentResult(success=True, called_agent_id=agent_tag)

    def _to_config(self) -> CallAgentConfig:
        """Convert current instance to config object"""
        return CallAgentConfig(known_agent_tags=self._known_agent_tags, sender_agent_id=self._sender_agent_id, group_chat_topic_type=self._group_chat_topic_type)

    @classmethod
    def _from_config(cls, config: CallAgentConfig) -> Self:
        """Create instance from config object"""
        return cls(known_agent_tags=config.known_agent_tags, sender_agent_id=config.sender_agent_id, group_chat_topic_type=config.group_chat_topic_type)



from typing import Optional

from autogen_core import CancellationToken, Component, ComponentModel
from autogen_core.code_executor import CodeExecutor
from autogen_core.tools import BaseTool
from pydantic import BaseModel, Field, model_serializer
from typing_extensions import Self

from spoox.environment._utils import output_truncat
from spoox.environment.code_executors.CodeExecutorLocal import CodeBlockTimeout


class CodeExecutionInput(BaseModel):
    code: str = Field(description="The Python code block that should be executed.")
    timeout: int = Field(
        description="Maximum duration (in seconds) the code may run. "
                    "Define only if you want to increase the default timeout of 20s. "
                    "Timeout value must not exceed 120s.",
        default=20
    )

class CodeExecutionResult(BaseModel):
    output: str
    exit_code: Optional[int] = None

    @model_serializer
    def ser_model(self) -> str:
        if self.exit_code is None:
            return f"<output> {self.output} </output>"
        return f"<exit-code> {self.exit_code} </exit-code>  \n  <output> {self.output} </output>"


class PythonToolConfig(BaseModel):
    """Configuration for PythonTool"""

    executor: ComponentModel
    output_max: int
    description: str = "Run a Python code block once in the users current directory."


class PythonTool(
    BaseTool[CodeExecutionInput, CodeExecutionResult], Component[PythonToolConfig]
):
    """
    A tool that executes Python code blocks in a code executor and returns output.

    Args:
        executor: The code executor that will be used to execute the code blocks.
    """

    component_config_schema = PythonToolConfig

    def __init__(self, executor: CodeExecutor, output_max: int = 20000):
        super().__init__(
            CodeExecutionInput,
            CodeExecutionResult,
            "PythonExecutor",
            "Run a Python code block once in the users current directory."
        )
        self._executor = executor
        self._output_max = output_max

    async def run(self, args: CodeExecutionInput, cancellation_token: CancellationToken) -> CodeExecutionResult:
        # execute code
        code_block = CodeBlockTimeout(code=args.code, language="python", timeout=args.timeout)
        result = await self._executor.execute_code_blocks(
            code_blocks=[code_block], cancellation_token=cancellation_token
        )
        # make sure the output is cut when too long
        output = output_truncat(result.output, self._output_max)
        return CodeExecutionResult(exit_code=result.exit_code, output=output)

    def _to_config(self) -> PythonToolConfig:
        """Convert current instance to config object"""
        return PythonToolConfig(executor=self._executor.dump_component(), output_max=self._output_max)

    @classmethod
    def _from_config(cls, config: PythonToolConfig) -> Self:
        """Create instance from config object"""
        executor = CodeExecutor.load_component(config.executor)
        return cls(executor=executor, output_max=config.output_max)



import subprocess
import tempfile
from pathlib import Path
from typing import List, Optional, ClassVar, Union

from autogen_core import CancellationToken, Component
from autogen_core.code_executor import CodeExecutor, CodeBlock, CodeResult
from pydantic import BaseModel


class CodeExecutorPersistentCommandLineConfig(BaseModel):
    """Configuration for PermanentCommandLineCodeExecutor"""

    timeout: int = 90
    work_dir: Optional[str] = None


class CodeExecutorPersistentCommandLine(CodeExecutor, Component[CodeExecutorPersistentCommandLineConfig]):
    component_config_schema = CodeExecutorPersistentCommandLineConfig

    SUPPORTED_LANGUAGES: ClassVar[List[str]] = [
        "bash",
        "shell",
        "sh",
    ]

    def __init__(self, timeout: int = 90, work_dir: Optional[Union[Path, str]] = Path.cwd()):
        super().__init__()

        if timeout < 1:
            raise ValueError("Timeout must be greater than or equal to 1.")
        self.timeout = timeout

        self._started = False
        self._work_dir = work_dir
        self._code_dir = None
        self._shell_exec = None
        self._runs_count = 0

    @property
    def runs_count(self) -> int:
        """Counter of executed code blocks."""
        return self._runs_count

    async def execute_code_blocks(self, code_blocks: List[CodeBlock],
                                  cancellation_token: CancellationToken) -> CodeResult:

        if len(code_blocks) != 1:
            raise RuntimeError(f"CodeExecutorPersistentCommandLine `code_blocks` must exactly contain one code block.")
        code_block = code_blocks[0]

        if not self._started:
            raise RuntimeError(f"CodeExecutorPersistentCommandLine must be started. Make sure `.start()` is called.")

        if code_block.language.lower() not in self.SUPPORTED_LANGUAGES:
            raise ValueError(f"Language '{code_block.language}' not supported by CodeExecutorPersistentCommandLine.")

        # write code block to the sh file
        code_block_file = Path(f"{self._code_dir.name}/code_block_{self._runs_count}.sh")
        code_block_file.write_text(f"#!/bin/bash\n{code_block.code}\n")
        code_block_file.chmod(0o755)

        # exec code file
        try:
            result = subprocess.run(
                [str(code_block_file)],
                capture_output=True,
                text=True,
                timeout=self.timeout
            )
        except subprocess.TimeoutExpired:
            std_aggr = "command execution failed - timeout error"
            return CodeResult(1, std_aggr)

        # format exec output
        if result.stdout and result.stderr:
            return CodeResult(result.returncode, f"<stdout>\n{result.stdout}\n</stdout>\n<stderr>\n{result.stderr}\n</stderr>")
        elif result.stdout:
            return CodeResult(result.returncode, result.stdout)
        elif result.stderr:
            return CodeResult(result.returncode, result.stderr)
        return CodeResult(result.returncode, "")

    async def start(self) -> None:
        if not self._started:
            self._code_dir = tempfile.TemporaryDirectory(dir=self._work_dir)
            self._setup_shell_exec()
            self._runs_count = 0
            self._started = True

    async def stop(self) -> None:
        if self._started:
            self._code_dir.cleanup()
            self._shell_exec.terminate()
            self._started = False

    async def restart(self) -> None:
        if self._started:
            await self.stop()
        await self.start()

    def _setup_shell_exec(self):
        self._shell_exec = subprocess.Popen(
            ['/bin/bash'],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            bufsize=1,
            cwd=self._work_dir
        )



from autogen_core import RoutedAgent, message_handler, MessageContext

from spoox.agents.UbuntuMASGroupAgent.Message import RequestToSpeak


class SMASStopperAgent(RoutedAgent):

    def __init__(self) -> None:
        super().__init__(description="The agent just stops the MAS flow and returns without taking any action.")

    @message_handler
    async def handle_request_to_speak(self, message: RequestToSpeak, ctx: MessageContext) -> None:
        # intentionally do nothing -> will stop runtime (see runtime.stop_when_idle in agent system)
        pass



import json
import pickle
import re
import sys
from pathlib import Path
from pprint import pprint

import questionary

from spoox.interface.LogInterface import LogInterface


if __name__ == "__main__":

    # scan current directory for result folders
    script_dir = Path(__file__).resolve().parent
    pattern = re.compile(r"^results_.*")
    result_dirs = []
    for entry in script_dir.iterdir():
        if entry.is_dir() and pattern.match(entry.name):
            result_dirs.append(entry.name)
    if not result_dirs:
        print("No benchmark result directories found.")
        sys.exit(0)

    # ask user to select benchmark results
    selected = questionary.select(
        "Select benchmark run results:",
        choices=result_dirs
    ).ask()
    selected_results_dir = script_dir / selected

    # load benchmark run meta-data
    with (selected_results_dir / f"ubuntu_bench_run_meta_data.pkl").open("rb") as f:
        meta_data = pickle.load(f)
    pprint(meta_data)

    # ask user to select benchmark run and task
    selected_run = questionary.select(
        "Select specific run:",
        choices=[str(r) for r in range(1, meta_data["runs"] + 1)]
    ).ask()
    selected_task = questionary.select(
        "Select specific task:",
        choices=[str(t) for t in meta_data["task_ids"]]
    ).ask()

    # print meta-data
    with (selected_results_dir / f"run_{selected_run}_task_{selected_task}_exec_meta_data.json").open("r") as f:
        meta_data = json.load(f)
        pprint(meta_data)

    # print chat history
    with (selected_results_dir / f"run_{selected_run}_task_{selected_task}_log_interface.pkl").open("rb") as f:
        log_interface: LogInterface = pickle.load(f)
        print_logging = bool(questionary.confirm("Print logs", default=True).ask())
        log_interface.print_all_logs(print_logging=print_logging)



import json
from abc import ABC, abstractmethod

from autogen_core import FunctionCall, CancellationToken
from autogen_core.models import FunctionExecutionResult

from spoox.interface.Interface import Interface

class Environment(ABC):

    def __init__(self):
        self.tools = []
        self._started = False
        self.additional_tool_descriptions = ""

        # special: CallAgentTool is set by the SMASSupervisorAgent
        self.call_agent_tool = None

    @abstractmethod
    async def start(self):
        pass

    @abstractmethod
    async def stop(self):
        pass

    @abstractmethod
    async def reset(self):
        pass

    @abstractmethod
    def get_tools(self, obj):
        pass

    @abstractmethod
    def get_additional_tool_descriptions(self, obj) -> [str]:
        pass

    async def execute_tool_call(
            self, tools, call: FunctionCall, cancellation_token: CancellationToken, interface: Interface, usage_stats: dict, caller_name: str = ""
    ) -> FunctionExecutionResult:

        # logging
        args_parsed = json.loads(call.arguments)
        tool_name = call.name
        interface.print_tool_call(
            f"**Tool**: {tool_name}  \n**Arguments**:  \n{args_parsed}  \n",
            f"{caller_name} - tool_call"
        )

        if not self._started:
            raise RuntimeError(f"Environment must be started. Make sure `.start()` is called.")

        # find tool by name and run it
        tool = next((tool for tool in tools if tool.name == call.name), None)
        if tool is None:
            feResult = FunctionExecutionResult(call_id=call.id, content=f"Tool '{tool_name}' is not known.", is_error=True, name=call.name)
        else:
            # run the tool and capture the result
            try:
                result = await tool.run_json(args_parsed, cancellation_token)
                feResult = FunctionExecutionResult(call_id=call.id, content=tool.return_value_as_string(result), is_error=False, name=tool.name)
            except Exception as e:
                feResult = FunctionExecutionResult(call_id=call.id, content=str(e), is_error=True, name=tool.name)

        # logging
        interface.print_tool_call(
            f"**Tool**: {feResult.name}  \n**Is error**: {feResult.is_error}  \n**Content**:  \n{feResult.content}  \n",
            f"{caller_name} - tool_call_result"
        )
        usage_stats['tool_calls'].append((call, feResult))
        if tool_name in usage_stats['tool_call_counts']:
            usage_stats['tool_call_counts'][tool_name] += 1
        else:
            usage_stats['tool_call_counts'][tool_name] = 1

        return feResult



import sys
from pathlib import Path

if __name__ == "__main__":

    home = Path("/home/linus")
    template_file = home / "template.md"

    # read in template
    if not template_file.exists() or not template_file.is_file():
        sys.exit("Found no template file.")
    with template_file.open() as f:
        content = f.read()

    # official awesome repo: https://github.com/sindresorhus/awesome

    # check if at least the main title is present
    if "Requirements for your pull request" not in content:
        sys.exit("The template is incorrect.")

    # commit 2f7d2c7 (2f7d2c71e6d7e0db8b99288779f78945d8b6c2b4) contained typos,
    # which were later fixed in 8e71414 (8e71414aea13959f183b193f0bb62b1034c44fb4);
    # the sentence "It should not contain the name of the list." was added in 2f7d2c7;
    # we are using those differences to check weather the correct commit (2f7d2c7) was pulled.
    if "It should not contain the name of the list." not in content:
        sys.exit("This is not the template of the requested commit.")
    if "Your entry here should include a short description about the project/theme of the list." not in content:
        sys.exit("This is not the template of the requested commit.")

    # check that all checkboxes are marked as completed
    if "- [ ]" in content:
        sys.exit("Not all checkboxes are marked as completed.")

    sys.exit(0)



import html

import questionary
from rich.console import Console
from rich.panel import Panel

from spoox.interface.Interface import Interface
from rich.markdown import Markdown
from enum import Enum


class CLIColor(Enum):
    DEFAULT = '#909090'
    ORANGE = '#ab3e03'
    DARKORANGE = '#70441a'
    GREY = '#555555'
    DARKGREY = '#3d3d3d'
    CYAN = '#008B8B'
    BLUE = '#193754'
    GREEN = '#2C5C27'
    LILA = '#2D1F5E'
    RED = '#521A20'


class CLInterface(Interface):

    console = Console()

    def __init__(self, logging_active: bool = False):
        super().__init__(logging_active)

    def print(self, out_text: str, title: str = "", color: CLIColor = CLIColor.DEFAULT) -> None:
        md = Markdown(html.escape(out_text))  # html.escape shows markdown with html tags
        panel = Panel(md, title=title, style=color.value)
        self.console.print(panel)

    def print_highlight(self, out_text: str, title: str = "") -> None:
        self.print(out_text, title, CLIColor.ORANGE)

    def print_shadow(self, out_text: str, title: str = "") -> None:
        self.print(out_text, title, CLIColor.GREY)

    def print_thought(self, out_text: str, title: str = "") -> None:
        self.print(out_text, title, CLIColor.CYAN)

    def print_command(self, out_text: str, title: str = "") -> None:
        self.print(out_text, title, CLIColor.BLUE)

    def print_tool_call(self, out_text: str, title: str = "") -> None:
        self.print(out_text, title, CLIColor.DARKORANGE)

    def print_user_message(self, out_text: str, title: str = "user input") -> None:
        self.print(out_text, title, CLIColor.LILA)

    def print_logging(self, out_text: str, title: str = "") -> None:
        if self.logging_active:
            self.print(out_text, title, CLIColor.DARKGREY)

    def request_user_input(self, query: str) -> str:
        questionary.print('')
        return questionary.text(query).ask()

    def request_select_choice(self, question: str, choices: [str]) -> str:
        questionary.print('')
        return questionary.select(question, choices).ask()

    def reset(self) -> None:
        pass



import re
import sys
from pathlib import Path

if __name__ == "__main__":

    home = Path("/home/linus")
    benchmark_copy_dir = home / "bench"
    number_of_tasks_sol_file = home / "number_of_tasks.txt"

    # count available tasks
    pattern = re.compile(r"^task_(\d+)_desc\.txt$")
    numbers = {
        int(match.group(1))
        for file in benchmark_copy_dir.iterdir()
        if (match := pattern.match(file.name))
    }

    # check 'number_of_tasks' solution file and validate content
    if not number_of_tasks_sol_file.is_file():
        sys.exit("Missing number_of_tasks.txt")
    with number_of_tasks_sol_file.open() as f:
        content = f.read().strip()
        if content != f"{len(numbers)}":
            sys.exit("Incorrect total number of tasks")

    sys.exit(0)



import uuid

from autogen_ext.tools.langchain import LangChainToolAdapter
from langchain_community.tools import DuckDuckGoSearchResults
from terminal_bench.terminal.tmux_session import TmuxSession

from spoox.environment.Environment import Environment
from spoox.environment.code_executors.CodeExecutorTerminalBench import CodeExecutorTerminalBench, \
    set_code_executor_tmux_session
from spoox.environment.tools.PythonTool import PythonTool
from spoox.environment.tools.ShellTool import ShellTool
from spoox.environment.tools.TerminalTool import TerminalTool, set_terminal_tmux_session


class TerminalBenchEnvironment(Environment):

    def __init__(self, session: TmuxSession):
        super().__init__()

        session_id = uuid.uuid4().hex[:8]
        set_terminal_tmux_session(session_id, session)
        set_code_executor_tmux_session(session_id, session)

        self._code_executor = CodeExecutorTerminalBench(session_id, timeout=2 * 60)

        self._shell_tool = ShellTool(self._code_executor)
        self._python_tool = PythonTool(self._code_executor)
        self._terminal_tool = TerminalTool(session_id)  # TerminalInteractiveProgramTool(session_id)
        self._search_tool = LangChainToolAdapter(DuckDuckGoSearchResults(output_format="list"))

    async def start(self):
        await self._code_executor.start()
        self._started = True

    async def stop(self):
        await self._code_executor.stop()
        self._started = False

    async def reset(self):
        await self._terminal_tool.reset()
        await self._code_executor.restart()

    def get_tools(self, obj):
        class_name = obj.__class__.__name__
        if class_name in ["SingletonAgent"]:
            return [self._shell_tool, self._python_tool, self._terminal_tool, self._search_tool]
        elif class_name in ["ExplorerAgent"]:
            return [self._shell_tool, self._terminal_tool, self._search_tool]
        elif class_name in ["SolverAgent", "SubTaskSolverAgent", "RefinerAgent"]:
            return [self._shell_tool, self._python_tool, self._terminal_tool, self._search_tool]
        elif class_name in ["TesterAgent"]:
            return [self._shell_tool, self._python_tool, self._terminal_tool]
        elif class_name in ["SMASSupervisorAgent"] and self.call_agent_tool is not None:
            return [self.call_agent_tool]
        # default: return no tools (e.g. Approver, Summarizer, ...)
        return []

    def get_additional_tool_descriptions(self, obj) -> [str]:
        class_name = obj.__class__.__name__
        shell_descr = f"""- Use Shell tool for simple, single bash commands. Do **not** use it for commands that open interactive terminal programs (e.g. git log or man)."""
        py_descr = f"""- Use PythonExecutor tool for complex logic, scripting, or data processing."""
        terminal_descr = f"""- Use Terminal tool to execute commands in a persistent terminal instance and get back the exact terminal screen. This is especially useful for interactive programs such as git log, man, or vim."""

        if class_name in ["SingletonAgent"]:
            return [shell_descr, py_descr, terminal_descr]
        elif class_name in ["ExplorerAgent"]:
            return [shell_descr, terminal_descr]
        elif class_name in ["SolverAgent", "SubTaskSolverAgent", "RefinerAgent"]:
            return [shell_descr, py_descr, terminal_descr]
        elif class_name in ["TesterAgent"]:
            return [shell_descr, py_descr, terminal_descr]
        # default: return no additional tool descriptions (e.g. Approver, Summarizer, ...)
        return []



class AgentError(Exception):
    """Base exception for all agent errors."""

    def __init__(self, agent_id: str, message: str):
        m = f"Agent exception - for agent: {agent_id} - {message}"
        super().__init__(m)
        self.agent_id = agent_id


class ModelClientError(AgentError):

    def __init__(self, agent_id: str, max_retrials: int, model_client_exception: Exception):
        super().__init__(agent_id,
                         f"Model client exception:\n{model_client_exception} and max reached ({max_retrials})")
        self.max_retrials = max_retrials
        self.model_client_exception = model_client_exception


class MaxOnlyTextMessagesError(AgentError):

    def __init__(self, agent_id: str, max_only_text_messages: int):
        super().__init__(
            agent_id, f"Max only text messages reached ({max_only_text_messages})")
        self.max_only_text_messages = max_only_text_messages


class MaxIterationsError(AgentError):

    def __init__(self, agent_id: str, max_iterations: int):
        super().__init__(agent_id, f"Max agent iterations reached ({max_iterations})")
        self.max_iterations = max_iterations


class MaxOllamaRetrialsError(AgentError):

    def __init__(self, agent_id: str, max_ollama_retrials: int):
        super().__init__(agent_id,
                         f"Max ollama response errors retrials reached ({max_ollama_retrials})")
        self.max_ollama_retrials = max_ollama_retrials



import sys
from pathlib import Path

if __name__ == "__main__":

    home = Path("/home/linus")
    code_file = home / "all_code.py"
    count_file = home / "function_count.txt"

    # check code file exists and read it
    if not count_file.is_file():
        sys.exit("Missing all_code.py")
    with code_file.open() as f:
        code_lines = f.readlines()

    # check that there is exactly one function definition
    func_lines = [line for line in code_lines if line.strip().startswith("def ")]
    if len(func_lines) != 1:
        sys.exit(f"Error: Expected 1 function definition, found {len(func_lines)}.")

    # check that all python code snippets were copied
    py_cp_t1 = any("def add(a, b)" in l for l in code_lines)
    py_cp_t2 = any("return a + b" in l for l in code_lines)
    py_cp_t3 = any("for i in range(5):" in l for l in code_lines)
    py_cp_t4 = any("print(i)" in l for l in code_lines)
    if not py_cp_t1 or not py_cp_t2 or not py_cp_t3 or not py_cp_t4:
        sys.exit(f"Not all python snippets were copied to all_code.py.")

    # make sure that only code snippets of html files were copied
    py_cp_t5 = any("if 5 > 4:" in l for l in code_lines)
    py_cp_t6 = any("print('comparison worked')" in l for l in code_lines)
    if py_cp_t5 or py_cp_t6:
        sys.exit(f"Some python snippets were copied to all_code.py that are not part of a html file.")

    # check function count file and validate content
    if not count_file.is_file():
        sys.exit("Missing function_count.txt")
    with count_file.open() as f:
        content = f.read().strip()
        if content != "1":  # Only one function defined across snippets
            sys.exit("Incorrect function count")

    sys.exit(0)





def output_truncat(output: str, output_max: int) -> str:
    if len(output) > output_max:
        info_text = "... **[The output exceeded the maximum allowed length, so the remaining/earlier content was truncated]** ..."
        return f"{info_text} {output[:output_max]}"
    return output



import os
import subprocess
import sys

if __name__ == "__main__":

    # attention: this file must work when executed with python 2.7 and python 3.x
    game_file = os.path.join("home", "linus", "simple_game", "game.py")
    game_file_original = "/opt/task_12_setup_py_program.py"

    # check if game python file's content was changed
    with open(game_file, 'rb') as f1, open(game_file_original, 'rb') as f2:
        while True:
            b1 = f1.read(4096)
            b2 = f2.read(4096)
            if b1 != b2:
                sys.exit("The game.py file's code was changed, that is not allowed.")
            if not b1:
                break

    # check if pyton game file can be executed
    try:
        proc = subprocess.Popen(
            ['python', game_file, '-i', '4'],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        stdout, stderr = proc.communicate()
        if proc.returncode != 0:
            print(stderr)
            sys.exit("Can not start the game.")
    except Exception as e:
        print("Error:", e)
        sys.exit("Can not start the game.")

    sys.exit(0)



import html
import subprocess

from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel

from spoox.interface.CLInterface import CLIColor
from spoox.interface.Interface import Interface


class LogInterfaceUserDelegate:

    def __init__(self):
        super().__init__()
        self._user_inputs: list = []
        self._user_choices: list = []
        self.default_user_input = ""
        self.default_user_choice = ""

    @property
    def user_input(self) -> str:
        if len(self._user_inputs) == 0:
            return self.default_user_input
        return self._user_inputs.pop(0)

    @property
    def user_choice(self) -> str:
        if len(self._user_choices) == 0:
            return self.default_user_choice
        return self._user_choices.pop(0)

    @user_input.setter
    def user_input(self, inp: str or [str]) -> None:
        if isinstance(inp, str):
            self._user_inputs.append(inp)
        if isinstance(inp, list):
            self._user_inputs += inp

    @user_choice.setter
    def user_choice(self, choice: str or [str]) -> None:
        if isinstance(choice, str):
            self._user_choices.append(choice)
        if isinstance(choice, list):
            self._user_choices += choice


class LogInterface(Interface):

    console = Console()

    def __init__(self, logging_active: bool = False, print_live: bool = False, feedback_iterations_max: int = 0, home_dir_path: str = None, eval_file_path: str = None):
        super().__init__(logging_active)

        if feedback_iterations_max > 0 and (eval_file_path is None or home_dir_path is None):
            raise ValueError("LogInterface: if feedback_iterations_max greater then 0, an eval_file_path is required.")

        self.logs = []
        self.user_delegate = LogInterfaceUserDelegate()
        self.feedback_iterations_max = feedback_iterations_max
        self.feedback_iterations_done = 0
        self.home_dir_path = home_dir_path
        self.eval_file_path = eval_file_path
        self.task_requested = False
        self.print_live = print_live

    def printCLI(self, out_text: str, title: str = "", color: CLIColor = CLIColor.DEFAULT) -> None:
        md = Markdown(html.escape(out_text))  # html.escape shows markdown with html tags
        panel = Panel(md, title=title, style=color.value)
        self.console.print(panel)

    def print(self, out_text: str, title: str = "", color: CLIColor = CLIColor.DEFAULT) -> None:
        if self.print_live:
            self.printCLI(out_text, title, color)
        self.logs.append((f"{title}", out_text, color.value))

    def print_highlight(self, out_text: str, title: str = "") -> None:
        self.print(out_text, title, CLIColor.ORANGE)

    def print_shadow(self, out_text: str, title: str = "") -> None:
        self.print(out_text, title, CLIColor.GREY)

    def print_thought(self, out_text: str, title: str = "") -> None:
        self.print(out_text, title, CLIColor.CYAN)

    def print_command(self, out_text: str, title: str = "") -> None:
        self.print(out_text, title, CLIColor.BLUE)

    def print_tool_call(self, out_text: str, title: str = "") -> None:
        self.print(out_text, title, CLIColor.DARKORANGE)

    def print_user_message(self, out_text: str, title: str = "user input") -> None:
        self.print(out_text, title, CLIColor.LILA)

    def print_logging(self, out_text: str, title: str = "") -> None:
        if self.logging_active:
            self.print(out_text, title, CLIColor.DARKGREY)

    def request_user_input(self, query: str) -> str:
        # if feedback_iterations_max > 0 we provide feedback in the form of the eval_file_path output when executed;
        # we do this max feedback_iterations_max times and only if eval_file_path was not successful.
        self.logs.append(("user_input_request", query))

        if not self.task_requested:
            self.task_requested = True
            user_input = self.user_delegate.user_input

        elif self.feedback_iterations_done < self.feedback_iterations_max:

            # execute eval script and collect output
            eval_success = False
            eval_error_stderr = ""
            try:
                result = subprocess.run(
                    ["python", str(self.eval_file_path)],
                    capture_output=True,
                    text=True,
                    check=True,
                    cwd=self.home_dir_path,
                    user='linus'
                )
                eval_success = result.returncode == 0
            except subprocess.CalledProcessError as e:
                eval_error_stderr = e.stderr or ""

            if eval_success:
                # eval successfully repeat with user_delegate
                user_input = self.user_delegate.user_input
            else:
                # pass feedback from the eval script
                user_input = eval_error_stderr
                self.feedback_iterations_done += 1

        else:
            user_input = self.user_delegate.user_input

        self.print_user_message(user_input)
        return user_input

    def request_select_choice(self, question: str, choices: [str]) -> str:
        self.logs.append(("select_choice_request_question", question))
        self.logs.append(("select_choice_request_choices", ', '.join(choices)))
        user_choice = self.user_delegate.user_choice
        self.logs.append(("selected_choice", user_choice))
        return user_choice

    def reset(self) -> None:
        self.logs = []
        self.user_delegate = LogInterfaceUserDelegate()

    def print_all_logs(self, print_logging: bool = False) -> None:
        """Print all self.logs."""

        for log in self.logs:
            if 'logging' in log[0] and not print_logging:
                continue
            md = Markdown(html.escape(log[1]))  # html.escape shows markdown with html tags
            panel = Panel(md, title=log[0], style=log[2] if len(log) > 2 else '#909090')
            self.console.print(panel)



import asyncio

from autogen_core.models import ChatCompletionClient

from spoox.agents.UbuntuMASGroupAgent.BaseGroupChatAgent import BaseGroupChatAgent
from spoox.agents.UbuntuMASGroupAgent.StructuredFlow.agents.utils import get_SUMMARIZER_SYSTEM_MESSAGE
from spoox.interface.Interface import Interface


class SummarizerAgent(BaseGroupChatAgent):

    def __init__(
            self,
            topic_type: str,
            group_chat_topic_type: str,
            model_client: ChatCompletionClient,
            interface: Interface,
            usage_stats: dict,
            return_next_time_possible_event: asyncio.Event
    ) -> None:

        super().__init__(
            group_chat_topic_type=group_chat_topic_type,
            description="Agent creating the final summary.",
            system_message=get_SUMMARIZER_SYSTEM_MESSAGE(topic_type),
            model_client=model_client,
            interface=interface,
            usage_stats=usage_stats,
            max_internal_iterations=10,
            return_next_time_possible_event=return_next_time_possible_event
        )



import asyncio
import uuid

from autogen_core import SingleThreadedAgentRuntime, DefaultTopicId
from autogen_core import TypeSubscription
from autogen_core.models import UserMessage, ChatCompletionClient

from spoox.agents.AgentSystem import AgentSystem
from spoox.agents.UbuntuMASGroupAgent.Message import GroupChatMessage, RequestToSpeak
from spoox.agents.UbuntuMASGroupAgent.StructuredFlow.agents.SolverAgent import SolverAgent
from spoox.agents.UbuntuMASGroupAgent.StructuredFlow.agents.SummarizerAgent import SummarizerAgent
from spoox.agents.UbuntuMASGroupAgent.StructuredFlow.agents.TesterAgent import TesterAgent
from spoox.environment.Environment import Environment
from spoox.interface.Interface import Interface


class UbuntuMASGroupChatSmall(AgentSystem):

    # all topic types
    group_chat_topic_type = "groupchat"
    solver_topic_type = "solver"
    tester_topic_type = "tester"
    summarizer_topic_type = "summarizer"

    def __init__(self, interface: Interface, model_client: ChatCompletionClient,
                 environment: Environment, timeout: int = 3600):

        super().__init__(interface, model_client, environment, timeout)

        self.runtime = SingleThreadedAgentRuntime()

        # create timeout event to tell agents to return next the time possible
        self._timeout_event = None

        # agents
        self._solver_agent = None
        self._tester_agent = None
        self._summarizer_agent = None

    async def build_mas(self):
        """setup all agents"""

        self._solver_agent = await SolverAgent.register(
            self.runtime,
            self.solver_topic_type,
            lambda: SolverAgent(
                topic_type=self.solver_topic_type,
                group_chat_topic_type=self.group_chat_topic_type,
                environment=self.environment,
                model_client=self.model_client,
                interface=self.interface,
                usage_stats=self.usage_stats,
                tester_agent_topic_type=self.tester_topic_type,
                return_next_time_possible_event=self._timeout_event,
            ),
        )
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=self.solver_topic_type, agent_type=self._solver_agent.type))
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=self.group_chat_topic_type, agent_type=self._solver_agent.type))

        self._tester_agent = await TesterAgent.register(
            self.runtime,
            self.tester_topic_type,
            lambda: TesterAgent(
                topic_type=self.tester_topic_type,
                group_chat_topic_type=self.group_chat_topic_type,
                environment=self.environment,
                model_client=self.model_client,
                interface=self.interface,
                usage_stats=self.usage_stats,
                previous_agent_topic_type=self.solver_topic_type,
                next_agent_topic_type=self.summarizer_topic_type,
                return_next_time_possible_event=self._timeout_event,
            ),
        )
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=self.tester_topic_type, agent_type=self._tester_agent.type))
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=self.group_chat_topic_type, agent_type=self._tester_agent.type))

        self._summarizer_agent = await SummarizerAgent.register(
            self.runtime,
            self.summarizer_topic_type,
            lambda: SummarizerAgent(
                topic_type=self.summarizer_topic_type,
                group_chat_topic_type=self.group_chat_topic_type,
                model_client=self.model_client,
                interface=self.interface,
                usage_stats=self.usage_stats,
                return_next_time_possible_event=self._timeout_event,
            ),
        )
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=self.summarizer_topic_type, agent_type=self._summarizer_agent.type))
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=self.group_chat_topic_type, agent_type=self._summarizer_agent.type))

    async def start(self):

        if self._timeout_event is None:
            self._timeout_event = asyncio.Event()
        await self.environment.start()
        await self.build_mas()

        # user input loop
        while True:

            user_input = self.interface.request_user_input("Query...")
            self.usage_stats['user_interactions_count'] += 1

            # exit
            if user_input in ['q', 'exit', 'stop']:
                break

            # reset stop signal
            if self._timeout_event.is_set():
                self._timeout_event.clear()

            # send user input to the group chat and trigger solver agent
            self.runtime.start()
            await self.runtime.publish_message(
                message=GroupChatMessage(nonce=str(uuid.uuid4()), body=UserMessage(content=user_input, source="User")),
                topic_id=DefaultTopicId(type=self.group_chat_topic_type)
            )
            await asyncio.sleep(0.1)  # ensuring the group msg can be observed before the RTS (I think it is not required - but not sure...)
            await self.runtime.publish_message(
                message=RequestToSpeak(nonce=str(uuid.uuid4())),
                topic_id=DefaultTopicId(type=self.solver_topic_type)
            )

            # wait until the agents are complete (runtime is idle)
            # if we just stop the runtime, but agents are still working on it -> autogen runtime will raise ValueError
            # therefore we use an event to signal all agent to stop the next time possible
            async def _timeout():
                await asyncio.sleep(self.timeout)
                error_message = "Agent System waiting for runtime.stop_when_idle timeout error"
                print(error_message)  # todo to be deleted
                self.interface.print_highlight(error_message, "TimeoutError")
                self.usage_stats["agent_errors"].append(("TimeoutError", error_message))
                self._timeout_event.set()
            timeout_task = asyncio.create_task(_timeout())
            await self.runtime.stop_when_idle()
            timeout_task.cancel()

        # stop entirely
        await self.environment.stop()
        await self.runtime.close()

    def init_usage_stats(self):
        self.usage_stats['user_interactions_count'] = 0
        self.usage_stats['llm_calls_count'] = 0
        self.usage_stats['tool_call_counts'] = dict()
        self.usage_stats['tool_calls'] = []
        self.usage_stats['ollama_response_error_count'] = 0
        self.usage_stats['model_client_exceptions'] = []
        self.usage_stats['agent_errors'] = []
        self.usage_stats['prompt_tokens'] = []
        self.usage_stats['completion_tokens'] = []
        self.usage_stats['next_agent_calling_chain'] = []
        self.usage_stats['group_chat_message_lengths'] = []

    def get_state(self):
        return {
            'solver_agent': self._solver_agent,
            'tester_agent': self._tester_agent,
            'summarizer_agent': self._summarizer_agent,
        }


import asyncio
import time
from typing import List

from autogen_core import RoutedAgent, message_handler, MessageContext, FunctionCall
from autogen_core.models import SystemMessage, LLMMessage, ChatCompletionClient, AssistantMessage, \
    FunctionExecutionResultMessage
from ollama import ResponseError

from spoox.agents.UbuntuSingletonAgent.Message import PublicMessage
from spoox.agents.errors import MaxOllamaRetrialsError, ModelClientError, MaxOnlyTextMessagesError, MaxIterationsError, \
    AgentError
from spoox.environment.Environment import Environment
from spoox.interface import Interface
from spoox.agents.UbuntuSingletonAgent.utils import get_SINGLETON_SYSTEM_PROMPT

# just in case the model is not using the tools or including the finished_tag
# we have to make sure that there is a limit of "only text messages"
MAX_ONLY_TEXT_MESSAGES = 3

MAX_OLLAMA_RESPONSE_ERRORS_RETRIALS = 5

MAX_MODEL_CLIENT_ERRORS_RETRIALS = 3


class SingletonAgent(RoutedAgent):

    finished_tag = "finished"

    def __init__(
            self,
            description: str,
            environment: Environment,
            model_client: ChatCompletionClient,
            interface: Interface,
            usage_stats: dict,
            return_next_time_possible_event: asyncio.Event,
            max_internal_iterations: int = 100,
    ) -> None:

        super().__init__(description=description)
        self._environment = environment
        self._model_client = model_client
        self._interface = interface
        self._usage_stats = usage_stats
        self._max_internal_iterations = max_internal_iterations
        self._return_next_time_possible_event = return_next_time_possible_event

        system_message = SystemMessage(content=get_SINGLETON_SYSTEM_PROMPT(self.finished_tag, environment.get_additional_tool_descriptions(self)))
        self._chat_history: List[LLMMessage] = [system_message]
        self._tools = environment.get_tools(self)

        for t in self._tools:
            self._interface.print_logging(str(t.schema), f"logging - {self.id.type} - tool_schema")
        self._interface.print_logging(system_message.content, f"logging - {self.id.type} - system_message")

    @message_handler
    async def handle_request_to_speak(self, message: PublicMessage, ctx: MessageContext) -> None:
        """
        if the agent is requested to speak, the llm is triggered;
        if the response includes the 'finished_tag' and no tool calls, the answer is printed and the agent exits;
        if the response contains tool calls, the tools are executed, and the llm is triggered again with the results;
        one of the tool calls could call a next agent, if so a RequestToSpeak message is posted.
        """

        # add the user message to session
        self._chat_history.append(message.body)

        # run agent loop; if agent loop fails, the agent simply not generates any response
        try:
            await self.agent_loop(ctx)
        except AgentError as e:
            self._interface.print_highlight(str(e), "Agent Error")
            self._usage_stats["agent_errors"].append((type(e).__name__, str(e)))
        except Exception as e:
            self._interface.print_highlight(str(e), "Unexpected Error")
            self._usage_stats["agent_errors"].append((type(e).__name__, str(e)))

    async def agent_loop(self, ctx: MessageContext):
        """Run llm over and over again until the agent is finished."""

        counter_only_text_messages = 0
        ollama_response_errors = 0
        model_client_errors = 0
        for i in range(1, self._max_internal_iterations + 1):

            # handling agent system timeout event
            if self._return_next_time_possible_event.is_set():
                return

            # request llm
            self._usage_stats['llm_calls_count'] += 1
            try:
                llm_res = await self._model_client.create(
                    messages=self._chat_history,
                    tools=self._tools,
                    cancellation_token=ctx.cancellation_token,
                )
            except ResponseError as e:
                ollama_response_errors += 1
                self._usage_stats['ollama_response_error_count'] += 1
                self._interface.print_highlight(str(e), "Ollama ResponseError")
                if ollama_response_errors > MAX_OLLAMA_RESPONSE_ERRORS_RETRIALS:
                    raise MaxOllamaRetrialsError(self.id.type, MAX_OLLAMA_RESPONSE_ERRORS_RETRIALS)
                else:
                    self._interface.print_shadow("Ollama ResponseError -> retry (MAX_OLLAMA_RESPONSE_ERRORS_RETRIALS not yet reached)", "Ollama ResponseError")
                    continue
            except Exception as e:
                model_client_errors += 1
                self._usage_stats['model_client_exceptions'].append(str(e))
                self._interface.print_highlight(str(e), "Model Client Error")
                if model_client_errors > MAX_MODEL_CLIENT_ERRORS_RETRIALS:
                    raise ModelClientError(self.id.type, MAX_MODEL_CLIENT_ERRORS_RETRIALS, e)
                else:
                    self._interface.print_shadow("Model client error -> retry (MAX_MODEL_CLIENT_ERRORS_RETRIALS not yet reached)", "Model Client Error")
                    start_time = time.time()
                    await asyncio.sleep(60)
                    print(f"short delay ofter model client exception: {(time.time() - start_time) / 60}")
                    continue
            self._usage_stats['prompt_tokens'].append(llm_res.usage.prompt_tokens)
            self._usage_stats['completion_tokens'].append(llm_res.usage.completion_tokens)
            ollama_response_errors = 0
            model_client_errors = 0

            # add the response to session
            self._chat_history.append(
                AssistantMessage(content=llm_res.content, thought=llm_res.thought, source=self.id.type))
            self._interface.print_logging(str(llm_res), f"logging - {self.id.type} - entire llm_res")

            # print thoughts if available
            if llm_res.thought:
                self._interface.print_thought(llm_res.thought, f"{self.id.type} - thought field")

            # check if just text
            if isinstance(llm_res.content, str):
                self._interface.print(llm_res.content, f"{self.id.type} - message")
                # check if `finished_tag` is included
                if f"[{self.finished_tag.lower()}]" in llm_res.content.lower():
                    return
                # check if MAX_ONLY_TEXT_MESSAGES is reached
                counter_only_text_messages += 1
                if counter_only_text_messages > MAX_ONLY_TEXT_MESSAGES:
                    raise MaxOnlyTextMessagesError(self.id.type, MAX_ONLY_TEXT_MESSAGES)
                continue

            # check if tool calls (if it is not string it has to be a list of tool calls)
            assert isinstance(llm_res.content, list) and all(
                isinstance(call, FunctionCall) for call in llm_res.content
            )
            counter_only_text_messages = 0
            # execute all tool calls and add results to session
            tool_results = await asyncio.gather(
                *[self._environment.execute_tool_call(self._tools, call, ctx.cancellation_token, self._interface, self._usage_stats, self.id.type)
                  for call in llm_res.content]
            )
            self._chat_history.append(FunctionExecutionResultMessage(content=tool_results))

        raise MaxIterationsError(self.id.type, self._max_internal_iterations)



import asyncio
import re
import time
import uuid
import copy
from typing import List

from autogen_core import RoutedAgent, message_handler, MessageContext, DefaultTopicId, FunctionCall
from autogen_core.models import SystemMessage, LLMMessage, UserMessage, ChatCompletionClient, AssistantMessage, \
    FunctionExecutionResultMessage
from ollama import ResponseError

from spoox.agents.UbuntuMASGroupAgent.Message import GroupChatMessage, RequestToSpeak
from spoox.agents.UbuntuMASGroupAgent.StructuredFlow.agents.utils import get_AGENT_FAILED_GROUP_CHAT_MESSAGE
from spoox.agents.errors import MaxOllamaRetrialsError, ModelClientError, MaxOnlyTextMessagesError, MaxIterationsError, \
    AgentError
from spoox.environment.Environment import Environment
from spoox.interface.Interface import Interface

# just in case the model is not using the tools or calling a next agent and only responses with a text
# we have to make sure that there is a limit of "only text messages"
MAX_ONLY_TEXT_MESSAGES = 3

MAX_OLLAMA_RESPONSE_ERRORS_RETRIALS = 5

MAX_MODEL_CLIENT_ERRORS_RETRIALS = 3


class BaseGroupChatAgent(RoutedAgent):

    def __init__(
            self,
            group_chat_topic_type: str,
            description: str,
            system_message: str,
            model_client: ChatCompletionClient,
            interface: Interface,
            usage_stats: dict,
            return_next_time_possible_event: asyncio.Event,
            environment: Environment = None,
            next_agent_topic_types: [str] = None,
            max_internal_iterations: int = 50,
            fallback_agent_topic_type: str = None,
            reset_on_request_to_speak: bool = False,
    ) -> None:

        super().__init__(description=description)
        self._group_chat_topic_type = group_chat_topic_type.lower()
        self._environment = environment
        self._model_client = model_client
        self._interface = interface
        self._usage_stats = usage_stats
        self._next_agent_topic_types = [n.lower() for n in next_agent_topic_types or []]
        self._max_internal_iterations = max_internal_iterations
        self._fallback_agent_topic_type = fallback_agent_topic_type.lower() if fallback_agent_topic_type else None
        self._return_next_time_possible_event = return_next_time_possible_event
        self._reset_on_request_to_speak = reset_on_request_to_speak

        self._chat_history: List[LLMMessage] = [SystemMessage(content=system_message)]
        self._chat_history_group_chat_only: List[LLMMessage] = [SystemMessage(content=system_message)]
        self._tools = environment.get_tools(self) if self._environment else []

        for t in self._tools:
            self._interface.print_logging(str(t.schema), f"logging - {self.id.type} - tool_schema")
        self._interface.print_logging(system_message, f"logging - {self.id.type} - system_message")

    @message_handler
    async def handle_group_chat_message(self, message: GroupChatMessage, ctx: MessageContext) -> None:
        """
        each agent keeps its own group chat history;
        therefore, it has to process every GroupChatMessage and keeps track of whom posted the message;
        own group chat messages are only stored in the _chat_history_group_chat_only, such that if the chat history
        is reset (if reset_on_request_to_speak), the fallback chat history does also contain own previous summaries.
        """
        self._chat_history_group_chat_only.extend(
            [
                UserMessage(content=f"Transferred to {message.body.source.capitalize()} agent.", source="system"),
                message.body,
            ]
        )
        if message.body.source != self.id.type:
            self._chat_history.extend(
                [
                    UserMessage(content=f"Transferred to {message.body.source.capitalize()} agent.", source="system"),
                    message.body,
                ]
            )

    @message_handler
    async def handle_request_to_speak(self, message: RequestToSpeak, ctx: MessageContext) -> None:
        """
        if the agent is requested to speak, the llm is triggered;
        after responding, the next selected agent is called by posting a new RequestToSpeak message.
        """

        # make sure the environment is reset (no influence of previous agents)
        if self._environment:
            await self._environment.reset()

        # reset chat history to group chat history only
        if self._reset_on_request_to_speak:
            self._chat_history = copy.deepcopy(self._chat_history_group_chat_only)
            self._interface.print_logging("reset to group chat history only on request to speak",
                                          f"logging - {self.id.type} - reset")

        # add to agents own history -> system message to adopt persona now
        self._chat_history.append(
            UserMessage(
                content=f"Transferred to {self.id.type.capitalize()} agent, adopt the persona immediately.",
                source="system"
            )
        )

        # logging
        logging_chat_hist = '| -> ' + ' -> '.join([str(h.content)[:40].replace('\n', '') for h in self._chat_history])
        self._interface.print_logging(logging_chat_hist, f"logging - {self.id.type} - chat history")

        # run agent loop; if agent loop fails, the agent simply not generates any response
        try:
            await self.agent_loop(ctx)
            return
        except AgentError as e:
            self._interface.print_highlight(str(e), "Agent Error")
            self._usage_stats["agent_errors"].append((type(e).__name__, str(e)))
        except Exception as e:
            self._interface.print_highlight(str(e), "Unexpected Error")
            self._usage_stats["agent_errors"].append((type(e).__name__, str(e)))

        # check if fallback
        if self._fallback_agent_topic_type:
            failure_message = get_AGENT_FAILED_GROUP_CHAT_MESSAGE(self.id.type, self._fallback_agent_topic_type)
            self._interface.print_shadow(failure_message)
            await self._send_group_chat_message(failure_message)
            await asyncio.sleep(
                0.1)  # ensuring the group msg can be observed before the RTS (I think it is not required - but not sure...)
            await self._send_request_to_speak(self._fallback_agent_topic_type)

        # if error and no fallback just return -> no next agent will be triggered -> autogen runtime exits

    async def agent_loop(self, ctx: MessageContext):
        """Run llm over and over again until the agent is finished."""

        counter_only_text_messages = 0
        ollama_response_errors = 0
        model_client_errors = 0
        for i in range(1, self._max_internal_iterations + 1):

            # handling agent system timeout event
            if self._return_next_time_possible_event.is_set():
                return

            # request llm
            self._usage_stats['llm_calls_count'] += 1
            try:
                llm_res = await self._model_client.create(
                    messages=self._chat_history,
                    tools=self._tools,
                    cancellation_token=ctx.cancellation_token,
                )
            except ResponseError as e:
                ollama_response_errors += 1
                self._usage_stats['ollama_response_error_count'] += 1
                self._interface.print_highlight(str(e), "Ollama ResponseError")
                if ollama_response_errors > MAX_OLLAMA_RESPONSE_ERRORS_RETRIALS:
                    raise MaxOllamaRetrialsError(self.id.type, MAX_OLLAMA_RESPONSE_ERRORS_RETRIALS)
                else:
                    self._interface.print_shadow(
                        "Ollama ResponseError -> retry (MAX_OLLAMA_RESPONSE_ERRORS_RETRIALS not yet reached)",
                        "Ollama ResponseError")
                    continue
            except Exception as e:
                model_client_errors += 1
                self._usage_stats['model_client_exceptions'].append(str(e))
                self._interface.print_highlight(str(e), "Model Client Error")
                if model_client_errors > MAX_MODEL_CLIENT_ERRORS_RETRIALS:
                    raise ModelClientError(self.id.type, MAX_MODEL_CLIENT_ERRORS_RETRIALS, e)
                else:
                    self._interface.print_shadow("Model client error -> retry (MAX_MODEL_CLIENT_ERRORS_RETRIALS not yet reached)", "Model Client Error")
                    start_time = time.time()
                    await asyncio.sleep(60)
                    print(f"short delay ofter model client exception: {(time.time() - start_time) / 60}")
                    continue
            self._usage_stats['prompt_tokens'].append(llm_res.usage.prompt_tokens)
            self._usage_stats['completion_tokens'].append(llm_res.usage.completion_tokens)
            ollama_response_errors = 0
            model_client_errors = 0

            # add the response to session
            self._chat_history.append(
                AssistantMessage(content=llm_res.content, thought=llm_res.thought, source=self.id.type))
            self._interface.print_logging(str(llm_res), f"logging - {self.id.type} - entire llm_res")

            # print thoughts if available
            if llm_res.thought:
                self._interface.print_thought(llm_res.thought, f"{self.id.type} - thought field")

            # check if list of tool calls
            if isinstance(llm_res.content, list) and all(
                    isinstance(call, FunctionCall) for call in llm_res.content) and self._environment:
                # execute all tool calls and add results to session
                counter_only_text_messages = 0
                tool_results = await asyncio.gather(
                    *[self._environment.execute_tool_call(self._tools, call, ctx.cancellation_token, self._interface,
                                                          self._usage_stats, self.id.type)
                      for call in llm_res.content]
                )
                self._chat_history.append(FunctionExecutionResultMessage(content=tool_results))
                # SMASSupervisorAgent special case: if tools contained an AgentCall execution -> exit
                if any(call.name == "CallAgent" for call in llm_res.content):
                    # special logging for SMAS agent system
                    call_agent_tool_calls = [t.arguments for t in llm_res.content if t.name == "CallAgent"]
                    self._usage_stats['supervisor_agent_calling_chain'].append(call_agent_tool_calls)
                    return
                # otherwise: trigger LLM again with tool results in _chat_history
                continue

            # check if just text (autogen: if it is not a list of tool calls, it has to be string)
            assert isinstance(llm_res.content, str)
            self._interface.print(llm_res.content, f"{self.id.type} - message")

            # check if agent finished and calls next agent
            for nt in self._next_agent_topic_types:
                patter = rf"\[[^\]]*{re.escape(nt)}[^\]]*\]"
                if re.search(patter, llm_res.content, flags=re.IGNORECASE):
                    # logging
                    self._usage_stats['next_agent_calling_chain'].append(nt)
                    # we assume that if an agent tag is included, this message contains the summary for the group chat
                    await self._send_group_chat_message(llm_res.content)
                    await asyncio.sleep(
                        0.1)  # ensuring the group msg can be observed before the RTS (I think it is not required - but not sure...)
                    await self._send_request_to_speak(nt)
                    return

            # check if no `_next_agent_topic_types` were defined; if so, the agent finishes if no tools are called
            if not self._next_agent_topic_types:
                await self._send_group_chat_message(llm_res.content)
                return

            # check if MAX_ONLY_TEXT_MESSAGES is reached
            counter_only_text_messages += 1
            if counter_only_text_messages > MAX_ONLY_TEXT_MESSAGES:
                raise MaxOnlyTextMessagesError(self.id.type, MAX_ONLY_TEXT_MESSAGES)

        raise MaxIterationsError(self.id.type, self._max_internal_iterations)

    async def _send_group_chat_message(self, message: str):
        self._usage_stats['group_chat_message_lengths'].append(len(message))
        await self.publish_message(
            message=GroupChatMessage(
                nonce=str(uuid.uuid4()),
                body=UserMessage(content=message, source=self.id.type)
            ),
            topic_id=DefaultTopicId(type=self._group_chat_topic_type),
        )

    async def _send_request_to_speak(self, agent_type: str):
        await self.publish_message(RequestToSpeak(nonce=str(uuid.uuid4())), DefaultTopicId(type=agent_type))



import argparse
import asyncio
import time
from pathlib import Path

import nest_asyncio
from dotenv import load_dotenv

from spoox.environment.LocalEnvironment import LocalEnvironment
from spoox.interface.LogInterface import LogInterface
from spoox.utils import setup_model_client, setup_agent_system, save_logs

nest_asyncio.apply()

"""
example usage:
python src/spoox/spoox_headless.py -m gpt-5-mini -a mas-group-chat-m -t "create an empty file named dodo in the current dir"
"""


def main() -> None:
    """Entry point for the spoox CLI."""

    parser = argparse.ArgumentParser(description="Spoox argument parser")
    parser.add_argument(
        "-m",
        "--model-id",
        required=False,
        default="gpt-5-mini",
        help="Model id (str)",
    )
    parser.add_argument(
        "-a",
        "--agent-id",
        required=False,
        default="singleton",
        help="Agent id (str)",
    )
    parser.add_argument(
        "-r",
        "--print-reasoning",
        required=False,
        default=True,
        help="Print reasoning process in terminal, default set to true (bool)",
    )
    parser.add_argument(
        "-d",
        "--in-docker",
        required=False,
        default=False,
        help="Should be set to true if called within a docker container and using Ollama model (bool)",
    )
    parser.add_argument(
        "-t",
        "--task",
        required=True,
        help="Task description (str)",
    )
    parser.add_argument(
        "-l",
        "--logs-dir",
        required=False,
        default="/tmp/spoox-logs",
        help="Logs dir path (str)",
    )

    args = parser.parse_args()

    model_id = str(args.model_id)
    agent_id = str(args.agent_id)
    print_reasoning = str(args.print_reasoning).lower() in ("yes", "true", "t", "y", "1")
    in_docker = str(args.in_docker).lower() in ("yes", "true", "t", "y", "1")
    task = str(args.task)
    logs_dir = Path(str(args.logs_dir))

    load_dotenv()

    # setup model client
    model_client = setup_model_client(model_id=model_id, docker_access=in_docker)

    # setup environment and interface
    environment = LocalEnvironment()

    # setup headless interface -> using log interface
    interface = LogInterface(
        logging_active=True,
        print_live=print_reasoning,
        feedback_iterations_max=0,  # just a placeholder; not supported by terminal bench
        eval_file_path=f"task_eval.py",  # just a placeholder; not supported by terminal bench
        home_dir_path='..',  # just a placeholder; not supported by terminal bench
    )
    interface.user_delegate.user_input = [task, 'q']
    interface.user_delegate.default_user_choice = 'confirm'

    # setup and run agent system
    agent = setup_agent_system(agent_id, model_client, environment, interface)
    start_time = time.time()
    try:
        asyncio.run(agent.start())
    except Exception as e:
        interface.print_highlight(str(e), f"Exception during agent system execution.")
    exec_minutes = (time.time() - start_time) / 60

    # saving logs
    logs_dir.mkdir(parents=True, exist_ok=True)
    save_logs(logs_dir, 0, '-', None, agent_id, model_id, None, None, exec_minutes, agent, interface, task)


if __name__ == "__main__":
    main()



import asyncio
import os
import re
import time
import uuid
from pathlib import Path

from dotenv import load_dotenv
from terminal_bench.agents import BaseAgent
from terminal_bench.agents.base_agent import AgentResult
from terminal_bench.agents.failure_mode import FailureMode
from terminal_bench.terminal.models import TerminalCommand
from terminal_bench.terminal.tmux_session import TmuxSession

from spoox.environment.TerminalBenchEnvironment import TerminalBenchEnvironment
from spoox.interface.LogInterface import LogInterface
from spoox.utils import setup_model_client, setup_agent_system, save_logs

"""
# run TerminalBench single task
tb run \
    --dataset terminal-bench-core==0.1.1 \
    --agent-import-path src.benchmark.terminal_bench.AgentuTerminalBench:AgentuTB \
    --task-id hello-world \
    --global-agent-timeout-sec 2000
    
# run TerminalBench entire bench
tb run \
    --dataset terminal-bench-core==0.1.1 \
    --agent-import-path src.benchmark.terminal_bench.AgentuTerminalBench:AgentuTB \
    --n-concurrent 2
    
# run 40 subset of TerminalBench 
tb run \
    --local-registry-path /Users/linus/Documents/TUM/MA/code/agentu/src/benchmark/terminal_bench/custom_registry_tb_first_half_40.json \
    --dataset terminal-bench-core-first-half-40==0.1.1 \
    --agent-import-path src.benchmark.terminal_bench.AgentuTerminalBench:AgentuTB \
    --global-agent-timeout-sec 2000 \
    --n-concurrent 2
    
# run 40 subset of TerminalBench - all default
tb run \
    --local-registry-path /Users/linus/Documents/TUM/MA/code/agentu/src/benchmark/terminal_bench/custom_registry_tb_first_half_40.json \
    --dataset terminal-bench-core-first-half-40==0.1.1 \
    --agent-import-path src.benchmark.terminal_bench.AgentuTerminalBench:AgentuTB
"""

_AGENT_ID = "mas-group-chat-m"  # "singleton",'mas-group-chat-s','mas-group-chat-m','mas-group-chat-l','mas-supervisor'
_MODEL_ID = "gpt-5-mini"  # "gpt-oss:20b","qwen3:14b","claude-sonnet-4-5","magistral:24b","gpt-5","gpt-5-mini"

class AgentuTB(BaseAgent):

    @staticmethod
    def name() -> str:
        return "agentu"

    def perform_task(
            self,
            instruction: str,
            session: TmuxSession,
            logging_dir: Path | None = None
    ) -> AgentResult:
        """
        Perform the task described by `task_description`.
        Args:
            instruction: The description of the task to perform.
            session: Optional tmux session to send commands to.
            logging_dir: The directory to optionally log your agent's output.
        Returns:
            An `AgentResult` object with the agent's token counts and optionally a
            failure mode for debugging.
        """

        # setup agent
        # setup interface
        load_dotenv()

        log_interface = LogInterface(
            logging_active=True,
            feedback_iterations_max=0,  # just a placeholder; not supported by terminal bench
            eval_file_path=f"task_eval.py",  # just a placeholder; not supported by terminal bench
            home_dir_path='.',  # just a placeholder; not supported by terminal bench
        )
        log_interface.user_delegate.user_input = [instruction, 'q']
        log_interface.user_delegate.default_user_choice = 'confirm'
        # setup environment
        model_client = setup_model_client(model_id=_MODEL_ID)
        environment = TerminalBenchEnvironment(session)
        # setup agent
        agent = setup_agent_system(_AGENT_ID, model_client, environment, log_interface)

        # uncomment for manual shell tool testing
        #asyncio.run(live_testing(environment, session))

        # unfortunately, most of the task containers do not have Python pre-installed,
        # however, the agent's python tool requires this,
        # therefore, we try to install it before triggering the agent
        session.send_command(TerminalCommand(command="apt update", block=True))
        session.send_command(TerminalCommand(command="apt install -y python3 python3-pip", block=True))
        session.send_command(TerminalCommand(command="ln -sf /usr/bin/python3 /usr/bin/python", block=True))
        session.send_command(TerminalCommand(command="ln -sf /usr/bin/pip3 /usr/bin/pip", block=True))
        session.send_command(TerminalCommand(command="python --version", block=True))
        session.send_command(TerminalCommand(command="pip --version", block=True))
        log_interface.print_shadow(session.capture_pane(), "PYTHON INSTALLATION")

        # run agent
        start_time = time.time()
        try:
            asyncio.run(agent.start())
            error = not agent.usage_stats['agent_errors']
        except Exception as e:
            log_interface.print_highlight(str(e), f"Exception during terminal bench agent system execution.")
            error = True
        exec_minutes = (time.time() - start_time) / 60

        # saving our own agentu logs
        current_dir = os.path.dirname(os.path.abspath(__file__))
        results_dir_name = f"results_tb_{_AGENT_ID}_{_MODEL_ID.replace(":", "-")}"
        results_dir_path = Path(current_dir) / results_dir_name
        results_dir_path.mkdir(parents=True, exist_ok=True)
        instruction_name = instruction[:40]
        instruction_name = re.sub(r'[^a-zA-Z0-9]', '_', instruction_name)
        instruction_name = re.sub(r'_+', '_', instruction_name)
        instruction_name = instruction_name.strip("_")
        # add a random id (cause descriptions can start with same sentences...)
        instruction_name += f"_{uuid.uuid4().hex[:4]}"
        save_logs(results_dir_path, 0, instruction_name, None, _AGENT_ID, _MODEL_ID,
                  None, None, exec_minutes, agent, log_interface, instruction)

        # fill out terminal bench AgentResult
        agent_result = AgentResult(
            total_input_tokens=sum(agent.usage_stats['prompt_tokens']),
            total_output_tokens=sum(agent.usage_stats['completion_tokens']),
            failure_mode=FailureMode.UNKNOWN_AGENT_ERROR if error else FailureMode.NONE,
        )

        session.stop()

        return agent_result



import json
import math
from pathlib import Path
from statistics import mean

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.cm import ScalarMappable
from matplotlib.colors import LinearSegmentedColormap, Normalize


def collect_and_concat_tb_logs_and_agentu_logs(run: str) -> dict:
    print(f"\n-- collecting and concatenating {run} logs")
    run_dir = Path.cwd().parent / run

    # get tb results data
    tb_results_dir = run_dir / [d for d in run_dir.iterdir() if d.is_dir() and d.name.endswith("_terminal_bench_logs")][0]
    tb_results_data = json.loads((tb_results_dir / "results.json").read_text())

    # collect tb tasks and logging data
    tb_data = dict()
    for result in tb_results_data['results']:
        tb_data[result['instruction']] = {
            'task_id': result['task_id'],
            'instruction': result['instruction'],
            'is_resolved': bool(result['is_resolved']),
            'failure_mode': result['failure_mode'],
            'total_input_tokens': int(result['total_input_tokens'] or 0),
            'total_output_tokens': int(result['total_output_tokens'] or 0)
        }
    resolved_count = len([d for d in tb_data.values() if d['is_resolved']])
    unresolved_count = len([d for d in tb_data.values() if not d['is_resolved']])
    print(f"-> tb results number of collected results: {len(tb_data.keys())} "
          f"(resolved: {resolved_count}; unresolved: {unresolved_count})")
    assert (resolved_count + unresolved_count) == len(tb_data.keys())

    # collect agentu logging data
    agentu_data = dict()
    for f in run_dir.iterdir():
        if f.is_file() and f.name.endswith("_exec_meta_data.json"):
            logs = json.loads(f.read_text())
            agentu_data[logs['task_instruction']] = logs
    print(f"-> agentu results number of collected results: {len(agentu_data.keys())}")
    assert len(agentu_data.keys()) == len(tb_data.keys())

    # concat both log dicts and use task_ids as keys and not instruction text
    logs_data = dict()
    matches = 0
    no_matches = 0
    for k, v in tb_data.items():
        logs_data[k] = v
        if k in agentu_data:
            matches += 1
            logs_data[k] = v | agentu_data[k]
        else:
            no_matches += 1
    logs_data = {v['task_id']: v for k, v in logs_data.items()}
    print(f"-> concatenated logs number of collected results: {len(logs_data.keys())}")
    print(f"-> concatenated logs: {matches} matches; {no_matches} no matches")
    assert len(logs_data.keys()) == len(tb_data.keys()) == len(agentu_data.keys())
    assert no_matches == 0
    return logs_data


def calc_stats(logs: dict) -> dict:

    stats = dict()

    # accuracy
    n = len(logs)
    solved_c = len([t for t in logs if t['is_resolved']])
    stats['accuracy'] = solved_c / n

    # mean_total_prompt_tokens
    mtpt = mean(sum(t['agent_usage_stats']['prompt_tokens']) for t in logs)
    stats['mean_total_prompt_tokens'] = mtpt

    # mean_total_input_tokens
    mtit = mean(sum(t['agent_usage_stats']['completion_tokens']) for t in logs)
    stats['mean_total_completion_tokens'] = mtit

    # mean_llm_calls
    mlc = mean(t['agent_usage_stats']['llm_calls_count'] for t in logs)
    stats['mean_llm_calls'] = mlc

    # mean_exec_time_min
    metm = mean(t['exec_time'] for t in logs)
    stats['mean_exec_time_min'] = metm

    return stats




def plot_comparing_multiple_runs_two_attributes(data: list[dict], attr_1: str, attr_2: str):
    """
    expected data example:
    data = [
        {"label": "Singleton", "accuracy": [5, 4, 6], "in_tokens": [50, 40, 60]},
        {"label": "MAS-S", "accuracy": [7, 6, 7.5], "in_tokens": [500, 400, 600]},
    ]
    attr_1 = "accuracy"
    attr_2 = "in_tokens"
    """

    xs = list(range(len(data)))
    labels = [d.get("label", str(i)) for i, d in enumerate(data)]
    fig, ax1 = plt.subplots()

    # --- compute means per experiment ---
    y1_means = [mean(d[attr_1]) for d in data]
    y2_means = [mean(d[attr_2]) for d in data]

    # ---- Accuracy scatter on left axis ----
    for i, d in enumerate(data):
        ax1.scatter([i] * len(d[attr_1]), d[attr_1], color="tab:blue", alpha=0.5, label=attr_1 if i == 0 else "",
                    marker="x")
    ax1.plot(xs, y1_means, marker="o", linestyle="-", markersize=4, label=f"{attr_1} mean",
             color="tab:blue")  # mean points + connecting line (left axis)
    ax1.set_ylabel(attr_1, color="tab:blue")
    ax1.tick_params(axis="y")

    # ---- Costs scatter on right axis ----
    ax2 = ax1.twinx()  # create second y-axis
    for i, d in enumerate(data):
        ax2.scatter([i] * len(d[attr_2]), d[attr_2], color="tab:orange", alpha=0.7, label=attr_2 if i == 0 else "",
                    marker="x")
    ax2.plot(xs, y2_means, marker="o", linestyle="-", markersize=4, label=f"{attr_2} mean",
             color="tab:orange")  # mean points + connecting line (left axis)
    ax2.set_ylabel(attr_2, color="tab:orange")
    ax2.tick_params(axis="y")

    # ---- Shared x-axis formatting ----
    ax1.set_xticks(xs)
    ax1.set_xticklabels(labels)
    ax1.set_title(f"{attr_1} vs {attr_2}")

    # ---- Legends ----
    lines1, labels1 = ax1.get_legend_handles_labels()
    lines2, labels2 = ax2.get_legend_handles_labels()
    ax1.legend(lines1 + lines2, labels1 + labels2, loc="lower right")

    fig.tight_layout()
    plt.show()


import math
from statistics import mean
import matplotlib.pyplot as plt

import math
from statistics import mean
import matplotlib.pyplot as plt

def plot_multi_comparing_runs_two_attributes(
        data: list[dict],
        attr_pairs: list[tuple[str, ...]],
        title: str | None = None,
        ylims: list[tuple[float | None, float | None]] | None = None,
):
    """
    Plot single- or dual-axis comparisons for given attribute pairs.

    Parameters
    ----------
    data : list of dict
        Each dict holds experiment results, e.g.
        {"label": "MAS-S", "accuracy": [7,6,7.5], "in_tokens": [500,400,600]}
    attr_pairs : list of tuples
        Tuples of attributes to compare.
        - ("a",)   -> single-axis plot for 'a'
        - ("a","b")-> dual-axis plot with left 'a', right 'b'
    title : str, optional
        Global figure title
    ylims : list of (ymin, ymax), optional
        One tuple per plot in attr_pairs.
        If None, no limits applied.
        Example: [(0,1), (None,1000), ...]
    """
    if not attr_pairs:
        raise ValueError("attr_pairs must contain at least one tuple, e.g. ('accuracy',) or ('accuracy','in_tokens').")

    if ylims is not None and len(ylims) != len(attr_pairs):
        raise ValueError("ylims must have the same length as attr_pairs")

    xs = list(range(len(data)))
    labels = [d.get("label", str(i)) for i, d in enumerate(data)]

    n = len(attr_pairs)
    ncols = 2
    nrows = math.ceil(n / ncols)

    fig, axes = plt.subplots(nrows, ncols, figsize=(5 * ncols, 4.2 * nrows), squeeze=False)

    for idx, pair in enumerate(attr_pairs):
        r, c = divmod(idx, ncols)
        ax1 = axes[r][c]

        ymin, ymax = (ylims[idx] if ylims else (None, None))

        if len(pair) == 1:
            # --- single axis ---
            attr_1 = pair[0]
            y1_means = [mean(d[attr_1]) for d in data]

            for i, dct in enumerate(data):
                ax1.scatter([i] * len(dct[attr_1]), dct[attr_1],
                            alpha=0.5, marker="x", color="tab:blue",
                            label=(attr_1 if i == 0 else ""))
            ax1.plot(xs, y1_means, marker="o", linestyle="--", markersize=4,
                     label=f"{attr_1} mean", color="tab:blue")

            ax1.set_ylabel(attr_1, color="tab:blue")
            ax1.tick_params(axis="y")
            if ymin is not None or ymax is not None:
                ax1.set_ylim(bottom=ymin, top=ymax)

            ax1.set_xticks(xs)
            ax1.set_xticklabels(labels)
            #ax1.set_title(attr_1)

            l1, lab1 = ax1.get_legend_handles_labels()
            if l1:
                ax1.legend(l1, lab1, loc="lower right", fontsize=9)

        elif len(pair) == 2:
            # --- dual axis ---
            attr_1, attr_2 = pair
            y1_means = [mean(d[attr_1]) for d in data]
            y2_means = [mean(d[attr_2]) for d in data]

            # left
            for i, dct in enumerate(data):
                ax1.scatter([i] * len(dct[attr_1]), dct[attr_1],
                            alpha=0.5, marker="x", color="tab:blue",
                            label=(attr_1 if i == 0 else ""))
            ax1.plot(xs, y1_means, marker="o", linestyle="--", markersize=4,
                     label=f"{attr_1} mean", color="tab:blue")
            ax1.set_ylabel(attr_1, color="tab:blue")
            ax1.tick_params(axis="y")
            if ymin is not None or ymax is not None:
                ax1.set_ylim(bottom=ymin, top=ymax)

            # right
            ax2 = ax1.twinx()
            for i, dct in enumerate(data):
                ax2.scatter([i] * len(dct[attr_2]), dct[attr_2],
                            alpha=0.5, marker="x", color="tab:orange",
                            label=(attr_2 if i == 0 else ""))
            ax2.plot(xs, y2_means, marker="o", linestyle="--", markersize=4,
                     label=f"{attr_2} mean", color="tab:orange")
            ax2.set_ylabel(attr_2, color="tab:orange")
            ax2.tick_params(axis="y")
            if ymin is not None or ymax is not None:
                ax2.set_ylim(bottom=ymin, top=ymax)

            ax1.set_xticks(xs)
            ax1.set_xticklabels(labels)
            #ax1.set_title(f"{attr_1} vs {attr_2}")

            l1, lab1 = ax1.get_legend_handles_labels()
            l2, lab2 = ax2.get_legend_handles_labels()
            if l1 or l2:
                ax1.legend(l1 + l2, lab1 + lab2, loc="lower right", fontsize=9)

        else:
            raise ValueError(f"Each tuple must have length 1 or 2, got: {pair}")

    # hide unused subplots
    for k in range(n, nrows * ncols):
        r, c = divmod(k, ncols)
        axes[r][c].axis("off")

    if title:
        fig.suptitle(title, y=0.995, fontsize=14)
    fig.tight_layout()
    plt.show()

def plot_task_correctness_colored_bars(data: dict[int, list[float]], vmin=0.0, vmax=1.0, edge=False):
    """
        Each key -> stacked bar.
          - Segments have height = value
          - Values sorted descending so high (green) at bottom, low (orange) at top
          - Color maps 1 -> green, 0 -> orange
        """
    # sort bars by total height
    items = sorted(data.items(), key=lambda kv: sum(kv[1]), reverse=True)

    x = np.arange(len(items))

    # colormap: 0 = orange, 1 = green
    cmap = LinearSegmentedColormap.from_list("orange_to_green", ["darkorange", "darkgreen"])
    norm = Normalize(vmin=vmin, vmax=vmax)

    fig, ax = plt.subplots()

    for xi, (key, values) in enumerate(items):
        bottom = 0
        # sort values descending → green at bottom, orange at top
        for v in sorted(values, reverse=True):
            vv_clip = np.clip(float(v), vmin, vmax)
            color = cmap(norm(vv_clip))
            ax.bar(
                xi, 2, bottom=bottom,  # fixed height=1 # todo should be 1 not 2 -> required cause of subset 40
                color=color if key != 0 else "lightgrey", edgecolor="black", linewidth=0.6 if edge else 0
            )
            bottom += 1

    # x labels
    ax.set_xticks(x)
    ax.set_xticklabels([str(k) for k, _ in items])
    ax.set_ylabel("tasks count")
    ax.set_title("task solver appropriately (1 = task solved by the appropriate agents | 0 = task not solved by the appropriate agents)")

    # colorbar
    sm = ScalarMappable(norm=norm, cmap=cmap)
    sm.set_array([])
    cbar = plt.colorbar(sm, ax=ax, pad=0.01)
    cbar.set_label("appropriate scoring")

    plt.tight_layout()
    plt.show()


import sys
from pathlib import Path

if __name__ == "__main__":

    home = Path("/home/linus")
    original_script = home / "lemon" / "tree" / "rand_n.py"
    copied_script = home / "rand_n.py"

    if not original_script.exists():
        sys.exit("Copy the script without moving the original.")
    if not copied_script.exists():
        sys.exit("There is no copy of the requested script.")

    if original_script.read_text() != copied_script.read_text():
        sys.exit("The script is not the same.")

    sys.exit(0)



import tempfile
import uuid
from pathlib import Path
from typing import List, ClassVar

from autogen_core import CancellationToken, Component
from autogen_core.code_executor import CodeExecutor, CodeBlock, CodeResult
from pydantic import BaseModel
from terminal_bench.terminal.models import TerminalCommand
from terminal_bench.terminal.tmux_session import TmuxSession


_TMUX_SESSION: dict[str, TmuxSession] = {}


def set_code_executor_tmux_session(key: str, session: TmuxSession) -> None:
    global _TMUX_SESSION
    _TMUX_SESSION[key] = session


def get_code_executor_tmux_session(key: str) -> TmuxSession:
    if key not in _TMUX_SESSION:
        raise ValueError(f"TerminalTool: no global Tmux session set for key {key}.")
    return _TMUX_SESSION[key]


class CodeExecutorTerminalBenchConfig(BaseModel):
    """Configuration for CodeExecutorTerminalBenchConfig"""

    tmux_session_key: str
    timeout: int = 90


class CodeExecutorTerminalBench(CodeExecutor, Component[CodeExecutorTerminalBenchConfig]):
    component_config_schema = CodeExecutorTerminalBenchConfig

    SUPPORTED_LANGUAGES: ClassVar[List[str]] = [
        "python",
        "bash",
    ]

    def __init__(self, tmux_session_key: str, timeout: int = 90):
        super().__init__()

        if timeout < 1:
            raise ValueError("Timeout must be greater than or equal to 1.")
        self.timeout = timeout

        self._tmux_session_key = tmux_session_key
        self._session = get_code_executor_tmux_session(tmux_session_key)
        self._started = False
        self._runs_count = 0
        self._local_code_dir = None
        self._local_code_dir_path = None
        self._container_code_dir_path = None

    @property
    def runs_count(self) -> int:
        """Counter of executed code blocks."""
        return self._runs_count

    async def execute_code_blocks(self, code_blocks: List[CodeBlock],
                                  cancellation_token: CancellationToken = None) -> CodeResult:

        if len(code_blocks) != 1:
            raise RuntimeError(f"CodeExecutorTerminalBench `code_blocks` must exactly contain one code block.")
        code_block = code_blocks[0]

        if not self._started:
            raise RuntimeError(f"CodeExecutorTerminalBench must be started. Make sure `.start()` is called.")

        if code_block.language.lower() not in self.SUPPORTED_LANGUAGES:
            raise ValueError(f"Language '{code_block.language}' not supported by CodeExecutorTerminalBench.")

        # setup code block language dependents
        type_of_file = ".py" if code_block.language.lower() == "python" else ".sh"
        caller_keyword = "python" if code_block.language.lower() == "python" else "bash"

        # write code block to the py or sh file and copy to TB docker container
        rand_id = uuid.uuid4().hex[:8]
        code_block_file = self._local_code_dir_path / f"code_block_{self._runs_count}_{rand_id}{type_of_file}"
        code_block_file.write_text(f"{code_block.code}\n")
        code_block_file.chmod(0o755)
        self._session.copy_to_container(
            code_block_file,
            container_dir=str(self._container_code_dir_path),
        )

        # run python file in container
        terminal_out, exit_code = self._send_command_get_output(
            f"{caller_keyword} {self._container_code_dir_path}/{code_block_file.name}"
        )

        # remove first and last line so that agent could not get confused
        # first line just contains calling the code file
        # last line which is just the typical input mask line
        terminal_out_lines = terminal_out.splitlines()
        if len(terminal_out_lines) >= 2:
            terminal_out_lines = terminal_out_lines[1:-1]
        terminal_out = "\n".join(terminal_out_lines)

        return CodeResult(exit_code, terminal_out)  # todo exit code 0 is not valid for all outputs...

    def _send_command_get_output(self, command: str) -> (str, int):
        """
        We can only retrieve the entire shell output, not just the last commands results.
        Therefore, we echo a marker before running the command to know where to cut.
        """
        marker = f"__MARKER-{self._runs_count}-{uuid.uuid4().hex[:8]}__"
        self._runs_count += 1
        self._session.send_command(TerminalCommand(command=f"echo '{marker}'", block=True))
        try:
            # adding `echo` ensures that we always have 3 lines
            # and can cut off the first and last one without removing the output
            self._session.send_command(
                TerminalCommand(command=f"{command} ; echo", max_timeout_sec=self.timeout, block=True))
        except TimeoutError:
            return f"Timeout error after {self.timeout} seconds. Be aware that bash commands which open interactive terminal interfaces (e.g. 'top', 'man' or 'vim') can also result in a timeout.", 124
        terminal_out = self._session.capture_pane(capture_entire=True)
        terminal_out_cleaned = terminal_out.split(marker)[-1].strip()
        return terminal_out_cleaned, -1

    async def start(self) -> None:
        if not self._started:
            # setup local temp code dir
            self._local_code_dir = tempfile.TemporaryDirectory(dir=Path.cwd())
            self._local_code_dir_path = Path(self._local_code_dir.name)
            # setup temp code dir in container
            self._container_code_dir_path = Path(f"/opt/{self._local_code_dir_path.name}")
            self._send_command_get_output(f"mkdir {self._container_code_dir_path}")
            # others
            self._runs_count = 0
            self._started = True

    async def stop(self) -> None:
        if self._started:
            self._local_code_dir.cleanup()
            self._started = False

    async def restart(self) -> None:
        if self._started:
            await self.stop()
        await self.start()

